# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3
# cython: initializedcheck=False

# Authors: Isak Samsten
# License: BSD 3 clause

cimport numpy as np
import numpy as np

from libc.math cimport NAN, sqrt, floor, INFINITY
from libc.stdlib cimport free, malloc, labs
from numpy cimport float64_t, intp_t, ndarray, double_t

from ..utils cimport _stats
from ..utils._misc cimport Heap, HeapElement

from copy import deepcopy

from ..utils cimport TSArray
from ..utils._misc cimport List
from ..utils._stats cimport (
    fast_mean_std,
    IncStats,
    inc_stats_init,
    inc_stats_add,
    inc_stats_remove,
    inc_stats_mean,
    inc_stats_variance,
)

from ..utils._parallel import run_in_parallel
from ..utils.validation import check_array


cdef class MetricList(List):

    cdef int reset(self, Py_ssize_t metric, TSArray X, TSArray Y) noexcept nogil:
        return (<Metric> self.get(metric)).reset(X, Y)

    cdef double distance(
        self,
        Py_ssize_t metric,
        TSArray X,
        Py_ssize_t x_index,
        TSArray Y,
        Py_ssize_t y_index,
        Py_ssize_t dim,
    ) noexcept nogil:
        return (<Metric> self.get(metric)).distance(X, x_index, Y, y_index, dim)

    cdef double _distance(
        self,
        Py_ssize_t metric,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len
    ) noexcept nogil:
        return (<Metric> self.get(metric))._distance(x, x_len, y, y_len)


cdef class SubsequenceMetricList(List):

    cdef int reset(self, Py_ssize_t metric, TSArray X) noexcept nogil:
        return (<SubsequenceMetric> self.get(metric)).reset(X)

    cdef int init_transient(
        self,
        Py_ssize_t metric,
        TSArray X,
        SubsequenceView *v,
        Py_ssize_t index,
        Py_ssize_t start,
        Py_ssize_t length,
        Py_ssize_t dim,
    ) noexcept nogil:
        return (<SubsequenceMetric> self.get(metric)).init_transient(
            X, v, index, start, length, dim
        )

    cdef int init_persistent(
        self,
        Py_ssize_t metric,
        TSArray X,
        SubsequenceView* v,
        Subsequence* s,
    ) noexcept nogil:
        return (<SubsequenceMetric> self.get(metric)).init_persistent(X, v, s)

    cdef int free_transient(self, Py_ssize_t metric, SubsequenceView *t) noexcept nogil:
        return (<SubsequenceMetric> self.get(metric)).free_transient(t)

    cdef int free_persistent(self, Py_ssize_t metric, Subsequence *t) noexcept nogil:
        return (<SubsequenceMetric> self.get(metric)).free_persistent(t)

    cdef int from_array(
        self,
        Py_ssize_t metric,
        Subsequence *s,
        object obj,
    ):
        return (<SubsequenceMetric> self.get(metric)).from_array(s, obj)

    cdef object to_array(
        self,
        Py_ssize_t metric,
        Subsequence *s
    ):
        return (<SubsequenceMetric> self.get(metric)).to_array(s)

    cdef double transient_distance(
        self,
        Py_ssize_t metric,
        SubsequenceView *v,
        TSArray X,
        Py_ssize_t index,
        Py_ssize_t *return_index,
    ) noexcept nogil:
        return (<SubsequenceMetric> self.get(metric)).transient_distance(
            v, X, index, return_index
        )

    cdef double persistent_distance(
        self,
        Py_ssize_t metric,
        Subsequence *s,
        TSArray X,
        Py_ssize_t index,
        Py_ssize_t *return_index,
    ) noexcept nogil:
        return (<SubsequenceMetric> self.get(metric)).persistent_distance(
            s, X, index, return_index
        )

    cdef Py_ssize_t transient_matches(
        self,
        Py_ssize_t metric,
        SubsequenceView *v,
        TSArray X,
        Py_ssize_t index,
        double threshold,
        double *distances,
        Py_ssize_t *indices,
    ) noexcept nogil:
        return (<SubsequenceMetric> self.get(metric)).transient_matches(
            v, X, index, threshold, distances, indices
        )

    cdef Py_ssize_t persistent_matches(
        self,
        Py_ssize_t metric,
        Subsequence *s,
        TSArray X,
        Py_ssize_t index,
        double threshold,
        double *distances,
        Py_ssize_t *indices,
    ) noexcept nogil:
        return (<SubsequenceMetric> self.get(metric)).persistent_matches(
            s, X, index, threshold, distances, indices
        )


cdef int _ts_view_update_statistics(
    SubsequenceView *v, const double* sample
) noexcept nogil:
    """Update the mean and standard deviation of a shapelet info struct """
    cdef double ex = 0
    cdef double ex2 = 0
    cdef Py_ssize_t i
    for i in range(v.length):
        current_value = sample[i]
        ex += current_value
        ex2 += current_value ** 2

    v.mean = ex / v.length
    ex2 = ex2 / v.length - v.mean * v.mean
    if ex2 > EPSILON:
        v.std = sqrt(ex2)
    else:
        v.std = 0.0
    return 0


cdef class SubsequenceMetric:

    cdef int reset(self, TSArray X) noexcept nogil:
        pass

    cdef int init_transient(
        self,
        TSArray X,
        SubsequenceView *v,
        Py_ssize_t index,
        Py_ssize_t start,
        Py_ssize_t length,
        Py_ssize_t dim,
    ) noexcept nogil:
        v.index = index
        v.dim = dim
        v.start = start
        v.length = length
        v.mean = NAN
        v.std = NAN
        v.extra = NULL
        return 0

    cdef int init_persistent(
        self, TSArray X, SubsequenceView* v, Subsequence* s
    ) noexcept nogil:
        s.dim = v.dim
        s.length = v.length
        s.mean = v.mean
        s.std = v.std
        s.ts_start = v.start
        s.ts_index = v.index
        s.extra = NULL
        s.data = <double*> malloc(sizeof(double) * v.length)
        if s.data == NULL:
            return -1

        cdef const double *sample = &X[v.index, v.dim, v.start]
        cdef Py_ssize_t i
        for i in range(v.length):
            s.data[i] = sample[i]
        return 0

    cdef int free_transient(self, SubsequenceView *v) noexcept nogil:
        return 0

    cdef int free_persistent(self, Subsequence *v) noexcept nogil:
        if v.data != NULL:
            free(v.data)
            v.data = NULL
        return 0

    cdef int from_array(self, Subsequence *v, object obj):
        dim, arr = obj
        v.dim = dim
        v.length = arr.shape[0]
        v.mean = NAN
        v.std = NAN
        v.data = <double*> malloc(v.length * sizeof(double))
        v.extra = NULL
        if v.data == NULL:
            return -1

        cdef Py_ssize_t i
        for i in range(v.length):
            v.data[i] = arr[i]
        return 0

    cdef object to_array(self, Subsequence *s):
        cdef Py_ssize_t j
        cdef ndarray[float64_t] arr = np.empty(s.length, dtype=float)
        with nogil:
            for j in range(s.length):
                arr[j] = s.data[j]

        return (s.dim, arr)

    cdef double transient_distance(
        self,
        SubsequenceView *v,
        TSArray X,
        Py_ssize_t index,
        Py_ssize_t *return_index=NULL,
    ) noexcept nogil:
        return self._distance(
            &X[v.index, v.dim, v.start],
            v.length,
            v.mean,
            v.std if v.std != 0 else 1.0,
            v.extra,
            &X[index, v.dim, 0],
            X.shape[2],
            return_index,
        )

    cdef double persistent_distance(
        self,
        Subsequence *v,
        TSArray X,
        Py_ssize_t index,
        Py_ssize_t *return_index=NULL,
    ) noexcept nogil:
        return self._distance(
            v.data,
            v.length,
            v.mean,
            v.std if v.std != 0 else 1.0,
            v.extra,
            &X[index, v.dim, 0],
            X.shape[2],
            return_index,
        )

    cdef Py_ssize_t transient_matches(
        self,
        SubsequenceView *v,
        TSArray X,
        Py_ssize_t index,
        double threshold,
        double *distances,
        Py_ssize_t *indicies,
    ) noexcept nogil:
        return self._matches(
            &X[v.index, v.dim, v.start],
            v.length,
            v.mean,
            v.std if v.std != 0 else 1.0,
            v.extra,
            &X[index, v.dim, 0],
            X.shape[2],
            threshold,
            distances,
            indicies,
        )

    cdef Py_ssize_t persistent_matches(
        self,
        Subsequence *s,
        TSArray X,
        Py_ssize_t index,
        double threshold,
        double *distances,
        Py_ssize_t *indicies,
    ) noexcept nogil:
        return self._matches(
            s.data,
            s.length,
            s.mean,
            s.std if s.std != 0 else 1.0,
            s.extra,
            &X[index, s.dim, 0],
            X.shape[2],
            threshold,
            distances,
            indicies,
        )

    cdef void transient_profile(
        self,
        SubsequenceView *s,
        TSArray x,
        Py_ssize_t i,
        double *dp,
    ) noexcept nogil:
        self._distance_profile(
            &x[s.index, s.dim, s.start],
            s.length,
            s.mean,
            s.std if s.std != 0 else 1.0,
            s.extra,
            &x[i, s.dim, 0],
            x.shape[2],
            dp,
        )

    cdef void persistent_profile(
        self,
        Subsequence *s,
        TSArray x,
        Py_ssize_t i,
        double *dp,
    ) noexcept nogil:
        self._distance_profile(
            s.data,
            s.length,
            s.mean,
            s.std if s.std != 0 else 1.0,
            s.extra,
            &x[i, s.dim, 0],
            x.shape[2],
            dp,
        )

    cdef void _distance_profile(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        double *dp,
    ) noexcept nogil:
        self._matches(
                s,
                s_len,
                s_mean,
                s_std,
                s_extra,
                x,
                x_len,
                INFINITY,
                dp,
                NULL,
            )

    cdef double _distance(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        Py_ssize_t *return_index=NULL,
    ) noexcept nogil:
        return NAN

    cdef Py_ssize_t _matches(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        double threshold,
        double *distances,
        Py_ssize_t *indicies,
    ) noexcept nogil:
        return 0


cdef class ScaledSubsequenceMetric(SubsequenceMetric):

    cdef int init_transient(
        self,
        TSArray X,
        SubsequenceView *v,
        Py_ssize_t index,
        Py_ssize_t start,
        Py_ssize_t length,
        Py_ssize_t dim,
    ) noexcept nogil:
        cdef int err = SubsequenceMetric.init_transient(
            self, X, v, index, start, length, dim
        )
        if err < 0:
            return err
        _ts_view_update_statistics(v, &X[v.index, v.dim, v.start])
        return 0

    cdef int from_array(
        self,
        Subsequence *v,
        object obj,
    ):
        cdef int err = SubsequenceMetric.from_array(self, v, obj)
        if err == -1:
            return -1

        dim, arr = obj
        v.mean = np.mean(arr)
        v.std = np.std(arr)
        if v.std <= EPSILON:
            v.std = 0.0
        return 0


cdef class ScaledSubsequenceMetricWrap(ScaledSubsequenceMetric):
    cdef Metric wrap
    cdef double *x_buffer
    cdef double *s_buffer

    def __cinit__(self, Metric wrap):
        self.wrap = wrap
        self.x_buffer = NULL
        self.s_buffer = NULL

    def __reduce__(self):
        return self.__class__, (self.wrap, )

    def __dealloc__(self):
        free(self.x_buffer)
        free(self.s_buffer)

    cdef int reset(self, TSArray X) noexcept nogil:
        if self.x_buffer != NULL:
            free(self.x_buffer)
        if self.s_buffer != NULL:
            free(self.s_buffer)

        self.x_buffer = <double*> malloc(sizeof(double) * X.shape[2])
        self.s_buffer = <double*> malloc(sizeof(double) * X.shape[2])
        return self.wrap.reset(X, X)

    cdef double _distance(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        Py_ssize_t *return_index=NULL,
    ) noexcept nogil:
        cdef double mean, std
        cdef double min_dist = INFINITY
        cdef Py_ssize_t i, j

        cdef IncStats stats
        inc_stats_init(&stats)

        for i in range(s_len - 1):
            self.s_buffer[i] = (s[i] - s_mean) / s_std
            inc_stats_add(&stats, 1.0, x[i])

        self.s_buffer[s_len - 1] = (s[s_len - 1] - s_mean) / s_std

        for i in range(x_len - s_len + 1):
            # add the last value
            inc_stats_add(
                &stats, 1.0, x[i + s_len - 1]
            )
            std = inc_stats_variance(&stats)
            if std == 0.0:
                std = 1.0
            else:
                std = sqrt(std)

            mean = stats.mean
            for j in range(s_len):
                self.x_buffer[j] = (x[i + j] - mean) / std

            # remove the first value from the mean/std
            inc_stats_remove(&stats, 1.0, x[i])

            if (
                self.wrap._eadistance(
                    self.s_buffer,
                    s_len,
                    self.x_buffer,
                    s_len,
                    &min_dist,
                )
            ):
                if return_index != NULL:
                    return_index[0] = i

        return min_dist

    cdef Py_ssize_t _matches(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        double threshold,
        double *distances,
        Py_ssize_t *indicies,
    ) noexcept nogil:
        cdef double mean, std
        cdef double tmp_dist
        cdef Py_ssize_t i, j, n

        cdef IncStats stats
        inc_stats_init(&stats)

        for i in range(s_len - 1):
            self.s_buffer[i] = (s[i] - s_mean) / s_std
            inc_stats_add(&stats, 1.0, x[i])

        self.s_buffer[s_len - 1] = (s[s_len - 1] - s_mean) / s_std

        n = 0
        for i in range(x_len - s_len + 1):
            # add the last value
            inc_stats_add(
                &stats, 1.0, x[i + s_len - 1]
            )
            std = inc_stats_variance(&stats)
            if std == 0.0:
                std = 1.0
            else:
                std = sqrt(std)

            mean = stats.mean
            for j in range(s_len):
                self.x_buffer[j] = (x[i + j] - mean) / std

            # remove the first value from the mean/std
            inc_stats_remove(&stats, 1.0, x[i])

            tmp_dist = threshold
            if (
                self.wrap._eadistance(
                    self.s_buffer,
                    s_len,
                    self.x_buffer,
                    s_len,
                    &tmp_dist,
                )
            ):
                if indicies != NULL:
                    indicies[n] = i

                distances[n] = tmp_dist
                n += 1

        return n


cdef class SubsequenceMetricWrap(SubsequenceMetric):
    cdef Metric wrap

    def __cinit__(self, Metric wrap):
        self.wrap = wrap

    def __reduce__(self):
        return self.__class__, (self.wrap, )

    cdef int reset(self, TSArray X) noexcept nogil:
        return self.wrap.reset(X, X)

    cdef double _distance(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        Py_ssize_t *return_index=NULL,
    ) noexcept nogil:
        cdef double min_dist = INFINITY
        cdef Py_ssize_t i, j

        for i in range(x_len - s_len + 1):
            if (
                self.wrap._eadistance(
                    s,
                    s_len,
                    x + i,
                    s_len,
                    &min_dist,
                )
            ):
                if return_index != NULL:
                    return_index[0] = i

        return min_dist

    cdef Py_ssize_t _matches(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        double threshold,
        double *distances,
        Py_ssize_t *indicies,
    ) noexcept nogil:
        cdef double tmp_dist
        cdef Py_ssize_t i, j, n

        n = 0
        for i in range(x_len - s_len + 1):
            tmp_dist = threshold
            if (
                self.wrap._eadistance(
                    s,
                    s_len,
                    x + i,
                    s_len,
                    &tmp_dist,
                )
            ):
                if indicies != NULL:
                    indicies[n] = i

                distances[n] = tmp_dist
                n += 1

        return n

cdef class Metric:

    cdef int reset(self, TSArray x, TSArray y) noexcept nogil:
        return 0

    cdef double distance(
        self,
        TSArray x,
        Py_ssize_t x_index,
        TSArray y,
        Py_ssize_t y_index,
        Py_ssize_t dim,
    ) noexcept nogil:
        return self._distance(
            &x[x_index, dim, 0],
            x.shape[2],
            &y[y_index, dim, 0],
            y.shape[2],
        )

    cdef double _distance(
        self,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len
    ) noexcept nogil:
        return NAN

    # Default implementation. Delegates to _eadistance.
    cdef bint eadistance(
        self,
        TSArray x,
        Py_ssize_t x_index,
        TSArray y,
        Py_ssize_t y_index,
        Py_ssize_t dim,
        double *distance,
    ) noexcept nogil:
        return self._eadistance(
           &x[x_index, dim, 0],
           x.shape[2],
           &y[y_index, dim, 0],
           y.shape[2],
           distance,
        )

    # Default implementation. Delegates to _distance,
    # without early abandoning.
    cdef bint _eadistance(
        self,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len,
        double *min_dist,
    ) noexcept nogil:
        cdef double dist = self._distance(x, x_len, y, y_len)
        if dist < min_dist[0]:
            min_dist[0] = dist
            return True
        else:
            return False

    @property
    def is_elastic(self):
        return False


cdef class CallableMetric(Metric):

    cdef object func

    def __init__(self, func):
        self.func = func

    cdef double _distance(
        self,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len
    ) noexcept nogil:
        cdef np.npy_intp x_shape[1]
        cdef np.npy_intp y_shape[1]
        x_shape[0] = <np.npy_intp> x_len
        y_shape[0] = <np.npy_intp> y_len
        with gil:
            try:
                return float(
                    self.func(
                        np.PyArray_SimpleNewFromData(1, x_shape, np.NPY_DOUBLE, x),
                        np.PyArray_SimpleNewFromData(1, y_shape, np.NPY_DOUBLE, y),
                    )
                )
            except:  # noqa: E722
                return NAN


cdef Py_ssize_t dilated_distance_profile(
    Py_ssize_t stride,
    Py_ssize_t dilation,
    Py_ssize_t padding,
    double *kernel,
    Py_ssize_t k_len,
    const double* x,
    Py_ssize_t x_len,
    Metric metric,
    double *x_buffer,
    double *k_buffer,
    double ea,
    double* out,
) noexcept nogil:
    cdef Py_ssize_t input_size = x_len + 2 * padding
    cdef Py_ssize_t kernel_size = (k_len - 1) * dilation + 1
    cdef Py_ssize_t output_size = <Py_ssize_t> floor((input_size - kernel_size) / stride) + 1

    cdef Py_ssize_t j  # the index in the kernel and input array
    cdef Py_ssize_t i  # the index of the output array
    cdef Py_ssize_t k  # buffer index
    cdef Py_ssize_t padding_offset
    cdef Py_ssize_t input_offset
    cdef Py_ssize_t kernel_offset
    cdef Py_ssize_t convolution_size
    cdef double tmp_dist
    cdef Py_ssize_t dp_size = 0

    for i in range(output_size):
        padding_offset = padding - i * stride
        if padding_offset > 0:
            if padding_offset % dilation == 0:
                kernel_offset = padding_offset
            else:
                kernel_offset = padding_offset + dilation - (padding_offset % dilation)
            input_offset = kernel_offset - padding_offset
        else:
            kernel_offset = 0
            input_offset = labs(padding_offset)

        convolution_size = (
            min(x_len, input_offset + kernel_size - max(0, padding_offset))
            - input_offset
        )
        k = 0
        for j from 0 <= j < convolution_size by dilation:
            x_buffer[k] = x[input_offset + j]
            k_buffer[k] = kernel[((j + kernel_offset) // dilation)]
            k += 1

        tmp_dist = ea
        if metric._eadistance(x_buffer, k, k_buffer, k, &tmp_dist):
            out[dp_size] = tmp_dist / (<float>k / k_len)
            dp_size += 1

    return dp_size


cdef Py_ssize_t scaled_dilated_distance_profile(
    Py_ssize_t stride,
    Py_ssize_t dilation,
    Py_ssize_t padding,
    double *kernel,
    Py_ssize_t k_len,
    const double* x,
    Py_ssize_t x_len,
    Metric metric,
    double *x_buffer,
    double *k_buffer,
    double ea,
    double* out,
) noexcept nogil:
    cdef Py_ssize_t input_size = x_len + 2 * padding
    cdef Py_ssize_t kernel_size = (k_len - 1) * dilation + 1
    cdef Py_ssize_t output_size = <Py_ssize_t> floor((input_size - kernel_size) / stride) + 1

    cdef Py_ssize_t j  # the index in the kernel and input array
    cdef Py_ssize_t i  # the index of the output array
    cdef Py_ssize_t k  # buffer index
    cdef Py_ssize_t padding_offset
    cdef Py_ssize_t input_offset
    cdef Py_ssize_t kernel_offset
    cdef Py_ssize_t convolution_size
    cdef double tmp_dist
    cdef double mean, std
    cdef Py_ssize_t dp_size = 0

    for i in range(output_size):
        padding_offset = padding - i * stride
        if padding_offset > 0:
            if padding_offset % dilation == 0:
                kernel_offset = padding_offset
            else:
                kernel_offset = padding_offset + dilation - (padding_offset % dilation)
            input_offset = kernel_offset - padding_offset
        else:
            kernel_offset = 0
            input_offset = labs(padding_offset)

        convolution_size = (
            min(x_len, input_offset + kernel_size - max(0, padding_offset))
            - input_offset
        )
        k = 0
        mean = 0
        std = 0
        for j from 0 <= j < convolution_size by dilation:
            x_buffer[k] = x[input_offset + j]
            k_buffer[k] = kernel[((j + kernel_offset) // dilation)]
            mean += x_buffer[k]
            std += x_buffer[k] * x_buffer[k]
            k += 1

        mean = mean / k_len
        std = std / k_len - mean * mean
        if std > EPSILON:
            std = sqrt(std)
        else:
            std = 1

        for j in range(k):
            x_buffer[j] = (x_buffer[j] - mean) / std

        tmp_dist = ea
        if metric._eadistance(x_buffer, k, k_buffer, k, &tmp_dist):
            out[dp_size] = tmp_dist / (<float>k / k_len)
            dp_size += 1

    return dp_size


cdef ndarray[intp_t] _new_match_array(Py_ssize_t *matches, Py_ssize_t n_matches):
    if n_matches > 0:
        match_array = np.empty(n_matches, dtype=np.intp)
        for i in range(n_matches):
            match_array[i] = matches[i]
        return match_array
    else:
        return None


cdef ndarray[double_t] _new_distance_array(double *distances, Py_ssize_t n_matches):
    if n_matches > 0:
        dist_array = np.empty(n_matches, dtype=np.double)
        for i in range(n_matches):
            dist_array[i] = distances[i]
        return dist_array
    else:
        return None


cdef class _PairwiseSubsequenceDistance:

    cdef TSArray X
    cdef Py_ssize_t[:, :] min_indices
    cdef double[:, :] distances,
    cdef Subsequence **shapelets
    cdef Py_ssize_t n_shapelets
    cdef SubsequenceMetric metric

    def __cinit__(
        self,
        double[:, :] distances,
        Py_ssize_t[:, :] min_indices,
        TSArray X,
        SubsequenceMetric metric
    ):
        self.X = X
        self.distances = distances
        self.min_indices = min_indices
        self.metric = metric
        self.shapelets = NULL
        self.n_shapelets = 0

    def __dealloc__(self):
        if self.shapelets != NULL:
            for i in range(self.n_shapelets):
                self.metric.free_persistent(self.shapelets[i])
                free(self.shapelets[i])
            free(self.shapelets)
            self.shapelets = NULL

    cdef void set_shapelets(self, list shapelets, Py_ssize_t dim):
        cdef Py_ssize_t i
        self.n_shapelets = len(shapelets)
        self.shapelets = <Subsequence**> malloc(sizeof(Subsequence*) * self.n_shapelets)
        cdef Subsequence *s
        for i in range(self.n_shapelets):
            s = <Subsequence*> malloc(sizeof(Subsequence))
            self.metric.from_array(s, (dim, shapelets[i]))
            self.shapelets[i] = s

    @property
    def n_work(self):
        return self.X.shape[0]

    def __call__(self, Py_ssize_t job_id, Py_ssize_t offset, Py_ssize_t batch_size):
        cdef Py_ssize_t i, j, min_index
        cdef Subsequence *s
        cdef double distance
        cdef SubsequenceMetric metric = deepcopy(self.metric)
        with nogil:
            metric.reset(self.X)
            for i in range(offset, offset + batch_size):
                for j in range(self.n_shapelets):
                    s = self.shapelets[j]
                    distance = metric.persistent_distance(
                        s, self.X, i, &min_index
                    )
                    self.distances[i, j] = distance
                    self.min_indices[i, j] = min_index


def _pairwise_subsequence_distance(
    list shapelets,
    TSArray x,
    int dim,
    SubsequenceMetric metric,
    n_jobs,
):
    n_samples = x.shape[0]
    distances = np.empty((n_samples, len(shapelets)), dtype=np.double)
    min_indicies = np.empty((n_samples, len(shapelets)), dtype=np.intp)
    subsequence_distance = _PairwiseSubsequenceDistance(
        distances,
        min_indicies,
        x,
        metric,
    )
    metric.reset(x)  # TODO: Move to _PairwiseSubsequenceDistance
    subsequence_distance.set_shapelets(shapelets, dim)
    run_in_parallel(subsequence_distance, n_jobs=n_jobs, require="sharedmem")
    return distances, min_indicies


def _paired_subsequence_distance(
    list shapelets,
    TSArray x,
    int dim,
    SubsequenceMetric metric
):
    cdef Py_ssize_t n_samples = x.shape[0]
    cdef Py_ssize_t i, min_index
    cdef double dist
    cdef Subsequence subsequence
    cdef double[:] distances = np.empty(n_samples, dtype=np.double)
    cdef Py_ssize_t[:] min_indices = np.empty(n_samples, dtype=np.intp)

    metric.reset(x)
    for i in range(n_samples):
        metric.from_array(&subsequence, (dim, shapelets[i]))
        with nogil:
            dist = metric.persistent_distance(&subsequence, x, i, &min_index)
            metric.free_persistent(&subsequence)
            distances[i] = dist
            min_indices[i] = min_index

    return distances.base, min_indices.base


def _subsequence_match(
    object y,
    TSArray x,
    double threshold,
    int dim,
    SubsequenceMetric metric,
    n_jobs,
):
    cdef Py_ssize_t n_samples = x.shape[0]
    cdef double *distances = <double*> malloc(sizeof(double) * x.shape[2])
    cdef Py_ssize_t *indicies = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * x.shape[2])
    cdef Py_ssize_t i, n_matches

    cdef list distances_list = []
    cdef list indicies_list = []

    cdef Subsequence subsequence
    metric.reset(x)
    metric.from_array(&subsequence, (dim, y))
    with nogil:
        for i in range(x.shape[0]):
            n_matches = metric.persistent_matches(
                &subsequence,
                x,
                i,
                threshold,
                distances,
                indicies,
            )
            with gil:
                indicies_list.append(_new_match_array(indicies, n_matches))
                distances_list.append(_new_distance_array(distances, n_matches))

    free(distances)
    free(indicies)

    metric.free_persistent(&subsequence)
    return indicies_list, distances_list


def _paired_subsequence_match(
    list y,
    TSArray x,
    double threshold,
    int dim,
    SubsequenceMetric metric,
    n_jobs,
):
    cdef Py_ssize_t n_samples = x.shape[0]
    cdef double *distances = <double*> malloc(sizeof(double) * x.shape[2])
    cdef Py_ssize_t *indices = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * x.shape[2])
    cdef Py_ssize_t i, n_matches

    cdef list distances_list = []
    cdef list indicies_list = []

    cdef Subsequence subsequence
    metric.reset(x)
    for i in range(n_samples):
        metric.from_array(&subsequence, (dim, y[i]))
        n_matches = metric.persistent_matches(
            &subsequence,
            x,
            i,
            threshold,
            distances,
            indices,
        )

        indicies_list.append(_new_match_array(indices, n_matches))
        distances_list.append(_new_distance_array(distances, n_matches))
        metric.free_persistent(&subsequence)

    free(distances)
    free(indices)

    return indicies_list, distances_list


cdef class _PairwiseDistanceBatch:
    cdef TSArray x
    cdef TSArray y
    cdef Py_ssize_t dim
    cdef Metric metric

    cdef double[:, :] distances

    def __init__(
        self,
        TSArray x,
        TSArray y,
        Py_ssize_t dim,
        Metric metric,
        double[:, :] distances
    ):
        self.x = x
        self.y = y
        self.dim = dim
        self.metric = metric
        self.distances = distances

    @property
    def n_work(self):
        return self.x.shape[0]

    def __call__(self, Py_ssize_t job_id, Py_ssize_t offset, Py_ssize_t batch_size):
        cdef Metric metric = deepcopy(self.metric)
        cdef Py_ssize_t i, j
        cdef Py_ssize_t y_samples = self.y.shape[0]

        with nogil:
            metric.reset(self.x, self.y)
            for i in range(offset, offset + batch_size):
                for j in range(y_samples):
                    self.distances[i, j] = metric.distance(
                        self.x, i, self.y, j, self.dim
                    )


def _pairwise_distance(
    TSArray x,
    TSArray y,
    Py_ssize_t dim,
    Metric metric,
    n_jobs,
):

    cdef double[:, :] out = np.empty((x.shape[0], y.shape[0]), dtype=float)
    run_in_parallel(
        _PairwiseDistanceBatch(
            x,
            y,
            dim,
            metric,
            out,
        ),
        n_jobs=n_jobs,
        require="sharedmem",
    )

    return out.base


cdef class _SingletonPairwiseDistanceBatch:
    cdef TSArray x
    cdef Py_ssize_t dim
    cdef Metric metric

    cdef double[:, :] distances

    def __init__(
        self,
        TSArray x,
        Py_ssize_t dim,
        Metric metric,
        double[:, :] distances
    ):
        self.x = x
        self.dim = dim
        self.metric = metric
        self.distances = distances

    @property
    def n_work(self):
        return self.x.shape[0]

    def __call__(self, Py_ssize_t job_id, Py_ssize_t offset, Py_ssize_t batch_size):
        cdef Metric metric = deepcopy(self.metric)
        cdef Py_ssize_t i, j
        cdef Py_ssize_t n_samples = self.x.shape[0]
        cdef double dist

        with nogil:
            metric.reset(self.x, self.x)
            for i in range(offset, offset + batch_size):
                for j in range(i + 1, n_samples):
                    dist = metric.distance(
                        self.x, i, self.x, j, self.dim
                    )
                    self.distances[i, j] = dist
                    self.distances[j, i] = dist


def _singleton_pairwise_distance(
    TSArray x,
    Py_ssize_t dim,
    Metric metric,
    n_jobs
):
    cdef Py_ssize_t n_samples = x.shape[0]
    cdef double[:, :] out = np.zeros((n_samples, n_samples), dtype=float)

    run_in_parallel(
        _SingletonPairwiseDistanceBatch(
            x,
            dim,
            metric,
            out,
        ),
        n_jobs=n_jobs,
        require="sharedmem",
    )
    return out.base


cdef class _ArgMinBatch:
    cdef double[:, :] distances
    cdef Py_ssize_t[:, :] indices

    cdef TSArray x
    cdef TSArray y
    cdef Metric metric
    cdef Py_ssize_t dim

    def __init__(
        self,
        TSArray x,
        TSArray y,
        Metric metric,
        Py_ssize_t dim,
        double[:, :] distances,
        Py_ssize_t[:, :] indices
    ):
        self.x = x
        self.y = y
        self.metric = metric
        self.dim = dim
        self.distances = distances
        self.indices = indices

    @property
    def n_work(self):
        return self.x.shape[0]

    def __call__(self, Py_ssize_t job_id, Py_ssize_t offset, Py_ssize_t batch_size):
        cdef double distance
        cdef Py_ssize_t i, j

        cdef Py_ssize_t k = self.distances.shape[1]
        cdef Py_ssize_t y_samples = self.y.shape[0]
        cdef Heap heap = Heap(k)
        cdef HeapElement e
        cdef Metric metric = deepcopy(self.metric)

        with nogil:
            metric.reset(self.x, self.y)
            for i in range(offset, offset + batch_size):
                distance = INFINITY
                heap.reset()

                for j in range(y_samples):
                    if metric.eadistance(self.x, i, self.y, j, self.dim, &distance):
                        heap.push(j, distance)

                        if heap.isfull():
                            distance = heap.maxvalue()
                        else:
                            distance = INFINITY

                for j in range(k):
                    e = heap.getelement(j)
                    self.indices[i, j] = e.index
                    self.distances[i, j] = e.value


def _argmin_distance(
    TSArray x,
    TSArray y,
    Py_ssize_t dim,
    Metric metric,
    Py_ssize_t k,
    n_jobs
):

    cdef Py_ssize_t x_samples = x.shape[0]
    cdef Py_ssize_t y_samples = y.shape[0]

    cdef Py_ssize_t[:, :] indices = np.zeros((x_samples, k), dtype=np.intp)
    cdef double[:, :] values = np.zeros((x_samples, k), dtype=float)

    run_in_parallel(
        _ArgMinBatch(
            x,
            y,
            metric,
            dim,
            values,
            indices,
        ),
        n_jobs=n_jobs,
        require="sharedmem"
    )

    return indices.base, values.base


cdef class _ArgminSubsequenceDistance:

    cdef double[:, :] distances
    cdef Py_ssize_t[:, :] indices

    cdef TSArray s
    cdef Py_ssize_t[:] s_len
    cdef TSArray t
    cdef Metric metric
    cdef Py_ssize_t dim

    def __cinit__(
        self,
        TSArray s,
        Py_ssize_t[:] s_len,
        TSArray t,
        Py_ssize_t dim,
        Metric metric,
        double[:, :] distances,
        Py_ssize_t[:, :] indices,
    ):
        self.s = s
        self.s_len = s_len
        self.t = t
        self.dim = dim
        self.metric = metric
        self.distances = distances
        self.indices = indices

    @property
    def n_work(self):
        return self.t.shape[0]

    def __call__(self, Py_ssize_t job_id, Py_ssize_t offset, Py_ssize_t batch_size):
        cdef double distance
        cdef Py_ssize_t i, j

        cdef Py_ssize_t k = self.distances.shape[1]
        cdef Heap heap = Heap(k)
        cdef HeapElement e
        cdef Metric metric = deepcopy(self.metric)

        with nogil:
            metric.reset(self.t, self.s)
            for i in range(offset, offset + batch_size):
                distance = INFINITY
                heap.reset()

                for j in range(self.t.shape[2] - self.s_len[i] + 1):
                    if (
                        metric._eadistance(
                            &self.s[i, self.dim, 0],
                            self.s_len[i],
                            &self.t[i, self.dim, j],
                            self.s_len[i],
                            &distance)
                    ):
                        heap.push(j, distance)

                    if heap.isfull():
                        distance = heap.maxvalue()
                    else:
                        distance = INFINITY

                for j in range(k):
                    e = heap.getelement(j)
                    self.indices[i, j] = e.index
                    self.distances[i, j] = e.value


cdef class _ScaledArgminSubsequenceDistance:

    cdef double[:, :] distances
    cdef Py_ssize_t[:, :] indices

    cdef TSArray s
    cdef Py_ssize_t[:] s_len
    cdef TSArray t
    cdef Metric metric
    cdef Py_ssize_t dim

    cdef double *s_buffer
    cdef double *t_buffer

    def __cinit__(
        self,
        TSArray s,
        Py_ssize_t[:] s_len,
        TSArray t,
        Py_ssize_t dim,
        Metric metric,
        double[:, :] distances,
        Py_ssize_t[:, :] indices,
    ):
        self.s = s
        self.s_len = s_len
        self.t = t
        self.dim = dim
        self.metric = metric
        self.distances = distances
        self.indices = indices
        self.s_buffer = <double*> malloc(sizeof(double) * s.shape[2])
        self.t_buffer = <double*> malloc(sizeof(double) * s.shape[2])

    def __dealloc__(self):
        free(self.s_buffer)
        free(self.t_buffer)

    @property
    def n_work(self):
        return self.t.shape[0]

    def __call__(self, Py_ssize_t job_id, Py_ssize_t offset, Py_ssize_t batch_size):
        cdef double distance
        cdef Py_ssize_t i, j, n
        cdef Py_ssize_t s_len
        cdef Py_ssize_t n_timestep = self.t.shape[2]
        cdef Py_ssize_t k = self.distances.shape[1]

        cdef Heap heap = Heap(k)
        cdef HeapElement e
        cdef Metric metric = deepcopy(self.metric)

        cdef double mean, std
        cdef IncStats stats
        with nogil:
            metric.reset(self.t, self.s)
            for i in range(offset, offset + batch_size):
                s_len = self.s_len[i]

                distance = INFINITY
                heap.reset()
                inc_stats_init(&stats)

                fast_mean_std(&self.s[i, self.dim, 0], s_len, &mean, &std)
                if std == 0.0:
                    std = 1.0

                for n in range(s_len - 1):
                    self.s_buffer[n] = (self.s[i, self.dim, n] - mean) / std
                    inc_stats_add(&stats, 1.0, self.t[i, self.dim, n])

                self.s_buffer[s_len - 1] = (
                    self.s[i, self.dim, s_len - 1] - mean
                ) / std

                for j in range(n_timestep - s_len + 1):
                    # add the last value
                    inc_stats_add(
                        &stats, 1.0, self.t[i, self.dim, j + s_len - 1]
                    )
                    std = inc_stats_variance(&stats)
                    if std == 0.0:
                        std = 1.0
                    else:
                        std = sqrt(std)

                    mean = stats.mean
                    for n in range(s_len):
                        self.t_buffer[n] = (self.t[i, self.dim, j + n] - mean) / std

                    # remove the first value from the mean/std
                    inc_stats_remove(&stats, 1.0, self.t[i, self.dim, j])

                    if (
                        metric._eadistance(
                            self.s_buffer,
                            s_len,
                            self.t_buffer,
                            s_len,
                            &distance)
                    ):
                        heap.push(j, distance)

                    if heap.isfull():
                        distance = heap.maxvalue()
                    else:
                        distance = INFINITY

                for j in range(k):
                    e = heap.getelement(j)
                    self.indices[i, j] = e.index
                    self.distances[i, j] = e.value

def _argmin_subsequence_distance(
    TSArray s,
    Py_ssize_t[:] s_len,
    TSArray x,
    Py_ssize_t dim,
    Metric metric,
    Py_ssize_t k,
    bint scaled=False,
    n_jobs=None,
):
    cdef Py_ssize_t n_samples = x.shape[0]
    assert s.shape[0] == n_samples

    cdef Py_ssize_t[:, :] indices = np.empty((n_samples, k), dtype=np.intp)
    cdef double[:, :] values = np.empty((n_samples, k), dtype=float)

    if scaled:
        Method = _ScaledArgminSubsequenceDistance
    else:
        Method = _ArgminSubsequenceDistance

    run_in_parallel(
        Method(
            s,
            s_len,
            x,
            dim,
            metric,
            values,
            indices,
        ),
        n_jobs=n_jobs,
        require="sharedmem",
    )

    return indices.base, values.base


cdef class _PairedDistanceBatch:
    cdef TSArray x
    cdef TSArray y
    cdef Py_ssize_t dim
    cdef Metric metric

    cdef double[:] distances

    def __init__(
        self,
        TSArray x,
        TSArray y,
        Py_ssize_t dim,
        Metric metric,
        double[:] distances,
    ):
        self.x = x
        self.y = y
        self.dim = dim
        self.metric = metric
        self.distances = distances

    @property
    def n_work(self):
        return self.x.shape[0]

    def __call__(self, Py_ssize_t job_id, Py_ssize_t offset, Py_ssize_t batch_size):
        cdef Metric metric = deepcopy(self.metric)
        cdef Py_ssize_t i
        with nogil:
            metric.reset(self.x, self.y)
            for i in range(offset, offset + batch_size):
                self.distances[i] = metric.distance(self.x, i, self.y, i, self.dim)


def _paired_distance(
    TSArray y,
    TSArray x,
    Py_ssize_t dim,
    Metric metric,
    n_jobs,
):
    cdef double[:] out = np.empty(y.shape[0], dtype=float)
    run_in_parallel(
        _PairedDistanceBatch(
            x,
            y,
            dim,
            metric,
            out,
        ),
        n_jobs=n_jobs,
        require="sharedmem",
    )

    return out.base


cdef class _DistanceProfile:
    cdef double[:, :] distance_profile
    cdef TSArray x
    cdef TSArray y
    cdef Py_ssize_t dim
    cdef SubsequenceMetric metric

    def __cinit__(
        self,
        TSArray y,
        TSArray x,
        Py_ssize_t dim,
        SubsequenceMetric metric,
        double[:, :] distance_profile,
    ):
        self.distance_profile = distance_profile
        self.x = x
        self.y = y
        self.metric = metric
        self.dim = dim

    @property
    def n_work(self):
        return self.x.shape[0]

    def __call__(self, Py_ssize_t job_id, Py_ssize_t offset, Py_ssize_t batch_size):
        cdef Py_ssize_t i, j
        cdef SubsequenceMetric metric = deepcopy(self.metric)
        cdef SubsequenceView view

        with nogil:
            metric.reset(self.x)
            for i in range(offset, offset + batch_size):
                metric.init_transient(self.y, &view, i, 0, self.y.shape[2], 0)
                metric._distance_profile(
                    &self.y[i, 0, 0],
                    self.y.shape[2],
                    view.mean,
                    view.std,
                    view.extra,
                    &self.x[i, 0, 0],
                    self.x.shape[2],
                    &self.distance_profile[i, 0],
                )
                metric.free_transient(&view)


def _distance_profile(
    TSArray y,
    TSArray x,
    Py_ssize_t dim,
    SubsequenceMetric metric,
    n_jobs=None
):
    cdef double[:, :] out = np.ones(
        (x.shape[0], x.shape[2] - y.shape[2] + 1), dtype=float
    )

    run_in_parallel(
        _DistanceProfile(
            y,
            x,
            dim,
            metric,
            out,
        ),
        n_jobs=n_jobs,
        require="sharedmem",
    )
    return out.base


cdef class _DilatedDistanceProfile:
    cdef TSArray S
    cdef TSArray X
    cdef Py_ssize_t dim
    cdef Metric metric
    cdef Py_ssize_t dilation
    cdef Py_ssize_t padding
    cdef bint scaled,
    cdef double[:, :] out
    cdef double *x_buffer
    cdef double *s_buffer

    def __cinit__(
        self,
        TSArray S,
        TSArray X,
        Py_ssize_t shapelet_size,
        Py_ssize_t dim,
        Metric metric,
        Py_ssize_t dilation,
        Py_ssize_t padding,
        bint scaled,
        double[:, :] out,
    ):
        self.S = S
        self.X = X
        self.dim = dim
        self.metric = metric
        self.dilation = dilation
        self.padding = padding
        self.scaled = scaled
        self.out = out
        self.x_buffer = <double*> malloc(sizeof(double) * shapelet_size)
        self.s_buffer = <double*> malloc(sizeof(double) * shapelet_size)

    def __dealloc__(self):
        free(self.x_buffer)
        free(self.s_buffer)

    @property
    def n_work(self):
        return self.X.shape[0]

    def __call__(self, Py_ssize_t job_id, Py_ssize_t offset, Py_ssize_t batch_size):
        cdef Metric metric = deepcopy(self.metric)
        cdef Py_ssize_t i

        with nogil:
            metric.reset(self.X, self.X)
            if self.scaled:
                for i in range(offset, offset + batch_size):
                    scaled_dilated_distance_profile(
                        1,
                        self.dilation,
                        self.padding,
                        &self.S[i, self.dim, 0],
                        self.S.shape[2],
                        &self.X[i, self.dim, 0],
                        self.X.shape[2],
                        metric,
                        self.x_buffer,
                        self.s_buffer,
                        INFINITY,
                        &self.out[i, 0],
                    )
            else:
                for i in range(offset, offset + batch_size):
                    dilated_distance_profile(
                        1,
                        self.dilation,
                        self.padding,
                        &self.S[i, self.dim, 0],
                        self.S.shape[2],
                        &self.X[i, self.dim, 0],
                        self.X.shape[2],
                        metric,
                        self.x_buffer,
                        self.s_buffer,
                        INFINITY,
                        &self.out[i, 0],
                    )


def _dilated_distance_profile(
    TSArray S,
    TSArray X,
    Py_ssize_t dim,
    Metric metric,
    Py_ssize_t dilation,
    Py_ssize_t padding,
    bint scaled,
    n_jobs,
):
    cdef Py_ssize_t shapelet_size = (S.shape[2] - 1) * dilation + 1
    cdef Py_ssize_t input_size = X.shape[2] + 2 * padding
    cdef double[:, :] out = np.empty(
        (X.shape[0], input_size - shapelet_size + 1), dtype=float
    )

    run_in_parallel(
        _DilatedDistanceProfile(
            S,
            X,
            shapelet_size,
            dim,
            metric,
            dilation,
            padding,
            scaled,
            out,
        ),
        n_jobs=n_jobs,
        require="sharedmem",
    )

    return out.base



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3
# cython: initializedcheck=False

# Authors: Isak Samsten
# License: BSD 3 clause
from libc.stdlib cimport free, malloc
from numpy cimport uint32_t

from ..utils cimport TSArray
from ..utils._misc cimport safe_realloc
from ..utils._rand cimport RAND_R_MAX
from ._attr_gen cimport Attribute, AttributeGenerator

from copy import deepcopy

import numpy as np
from joblib import Parallel, delayed, effective_n_jobs


def clone_embedding(AttributeGenerator generator, attributes):
    cdef AttributeEmbedding embedding = AttributeEmbedding(generator, len(attributes))
    cdef Py_ssize_t i
    cdef Attribute *attribute
    for i in range(len(attributes)):
        attribute = <Attribute*> malloc(sizeof(Attribute))
        generator.persistent_from_object(attributes[i], attribute)
        embedding.set_attribute(i, attribute)
    return embedding


def _partition_attributes(n_jobs, n_attributes, generator):
    n_jobs = min(effective_n_jobs(n_jobs), n_attributes)

    batch_size = n_attributes // n_jobs
    overflow = n_attributes % n_jobs
    generators = []
    attribute_offsets = []
    current_offset = 0
    batch_sizes = []
    for i in range(n_jobs):
        generators.append(deepcopy(generator))
        current_overflow = 0
        if i < overflow:
            current_overflow = 1
        current_batch_size = batch_size + current_overflow
        attribute_offsets.append(current_offset)
        batch_sizes.append(current_batch_size)
        current_offset += current_batch_size

    return n_jobs, generators, attribute_offsets, batch_sizes


cdef class Batch:

    cdef list generators
    cdef AttributeEmbedding embedding
    cdef double[:, :] x_out
    cdef TSArray x_in

    def __init__(self, list generators, AttributeEmbedding embedding):
        self.generators = generators
        self.embedding = embedding

    cdef void init(self, TSArray x_in, double[:, :] x_out):
        self.x_in = x_in
        self.x_out = x_out

cdef class BatchTransform(Batch):

    def __call__(self, Py_ssize_t job_id, Py_ssize_t attribute_offset, Py_ssize_t batch_size):
        cdef Py_ssize_t i, j
        cdef AttributeGenerator generator = self.generators[job_id]
        generator.reset(self.x_in)
        with nogil:
            for i in range(self.x_in.shape[0]):
                for j in range(batch_size):
                    generator.persistent_fill(
                        self.embedding.get_attribute(attribute_offset + j),
                        self.x_in,
                        i,
                        self.x_out,
                        i,
                        attribute_offset + j,
                    )

cdef class BatchFitTransform(Batch):

    def __call__(self, Py_ssize_t job_id, Py_ssize_t attribute_offset, Py_ssize_t batch_size, uint32_t seed):
        cdef Py_ssize_t i, j
        cdef AttributeGenerator generator = self.generators[job_id]
        generator.reset(self.x_in)
        cdef Py_ssize_t *samples = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * self.x_in.shape[0])
        cdef Attribute transient
        cdef Attribute *persistent

        with nogil:
            for i in range(self.x_in.shape[0]):
                samples[i] = i

            for j in range(batch_size):
                generator.next_attribute(
                    attribute_offset + j,
                    self.x_in,
                    samples,
                    self.x_in.shape[0],
                    &transient,
                    &seed,
                )
                for i in range(self.x_in.shape[0]):
                    generator.transient_fill(
                        &transient,
                        self.x_in,
                        i,
                        self.x_out,
                        i,
                        attribute_offset + j,
                    )
                persistent = <Attribute*> malloc(sizeof(Attribute))
                generator.init_persistent(
                    self.x_in, &transient, persistent
                )
                self.embedding.set_attribute(attribute_offset + j, persistent)
                generator.free_transient(&transient)

cdef class AttributeEmbedding:

    cdef AttributeGenerator generator
    cdef Attribute** _attributes
    cdef Py_ssize_t _n_attributes

    def __cinit__(self, AttributeGenerator generator, Py_ssize_t n_attributes):
        self.generator = generator
        self._attributes = <Attribute**> malloc(sizeof(Attribute*) * n_attributes)
        self._n_attributes = n_attributes

    def __dealloc__(self):
        cdef Py_ssize_t i
        for i in range(self._n_attributes):
            self.generator.free_persistent(self._attributes[i])
            free(self._attributes[i])
        free(self._attributes)

    def __reduce__(self):
        return clone_embedding, (
            self.generator,
            self.attributes,
        )

    cdef Py_ssize_t set_attribute(self, Py_ssize_t i, Attribute *attribute) noexcept nogil:
        """Add a attribute to the embedding

        attribute : Attribute*
            The attribute.

            Note: The attribute must be allocated using malloc
        """
        self._attributes[i] = attribute
        return 0

    cdef Attribute* get_attribute(self, Py_ssize_t i) noexcept nogil:
        return self._attributes[i]

    @property
    def n_features(self):
        import warnings
        warnings.warn(
            "`n_features` has been renamed to `n_attributes` in 1.2 and will be removed in 1.4",
            DeprecationWarning
        )
        return self.n_attributes

    @property
    def n_attributes(self):
        return self._n_attributes

    @property
    def features(self):
        import warnings
        warnings.warn(
            "`n_features` has been renamed to `n_attributes` in 1.2 and will be removed in 1.4",
            DeprecationWarning
        )
        return self.attributes

    @property
    def attributes(self):
        return [
            self.generator.persistent_to_object(self._attributes[i])
            for i in range(self._n_attributes)
        ]

    def __getitem__(self, item):
        if not isinstance(item, int) or 0 > item > self._n_attributes:
            raise ValueError()
        return self.generator.persistent_to_object(self._attributes[item])


def fit(AttributeGenerator generator, TSArray X, object random_state):
    cdef Py_ssize_t i
    cdef Attribute transient
    cdef Attribute *persistent
    cdef Py_ssize_t *samples = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * X.shape[0])
    generator.reset(X)
    cdef AttributeEmbedding embedding = AttributeEmbedding(
        generator, generator.get_n_attributess(X),
    )
    cdef uint32_t seed = random_state.randint(0, RAND_R_MAX)

    with nogil:
        for i in range(X.shape[0]):
            samples[i] = i

        for i in range(generator.get_n_attributess(X)):
            persistent = <Attribute*> malloc(sizeof(Attribute))
            generator.next_attribute(
                i,
                X,
                samples,
                X.shape[0],
                &transient,
                &seed,
            )
            generator.init_persistent(
                X, &transient, persistent
            )
            generator.free_transient(&transient)
            embedding.set_attribute(i, persistent)

    return embedding


def transform(AttributeEmbedding embedding, TSArray X, n_jobs=None):
    cdef AttributeGenerator generator = embedding.generator
    cdef Py_ssize_t n_outputs = generator.get_n_outputs(X)
    cdef Py_ssize_t n_attributes = generator.get_n_attributess(X)
    cdef double[:, :] out = np.empty((X.shape[0], n_outputs))

    n_jobs, generators, attribute_offsets, batch_sizes = _partition_attributes(
        n_jobs, n_attributes, generator
    )
    cdef BatchTransform transform = BatchTransform(generators, embedding)
    transform.init(X, out)

    Parallel(n_jobs=n_jobs, require="sharedmem")(
        delayed(transform)(
            jobid, attribute_offsets[jobid], batch_sizes[jobid]
        )
        for jobid in range(n_jobs)
    )

    return out.base


def fit_transform(AttributeGenerator generator, TSArray X, random_state, n_jobs=None):
    cdef Py_ssize_t n_outputs = generator.get_n_outputs(X)
    cdef Py_ssize_t n_attributes = generator.get_n_attributess(X)
    cdef AttributeEmbedding embedding = AttributeEmbedding(generator, n_attributes)
    cdef double[:, :] out = np.empty((X.shape[0], n_outputs))

    n_jobs, generators, attribute_offsets, batch_sizes = _partition_attributes(
        n_jobs, n_attributes, generator
    )
    seeds = random_state.randint(0, RAND_R_MAX, size=n_jobs)

    cdef BatchFitTransform fit_transform = BatchFitTransform(generators, embedding)
    fit_transform.init(X, out)
    Parallel(n_jobs=n_jobs, require="sharedmem")(
        delayed(fit_transform)(
            jobid, attribute_offsets[jobid], batch_sizes[jobid], seeds[jobid]
        )
        for jobid in range(n_jobs)
    )

    return embedding, out.base


def derivative_transform(TSArray X):
    if X.shape[2] < 3:
        return X.base

    cdef Py_ssize_t i, j, k
    cdef double[:, :, :] out = np.empty(
        (X.shape[0], X.shape[1], X.shape[2] - 2), dtype=float
    )

    for i in range(X.shape[0]):
        for k in range(X.shape[1]):
            for j in range(1, X.shape[2] - 1):
                out[i, k, j - 1] = (
                    (X[i, k, j] - X[i, k, j - 1])
                    + ((X[i, k, j + 1] - X[i, k, j - 1]) / 2)
                ) / 2

    return out.base



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3
# cython: initializedcheck=False

# Authors: Isak Samsten
# License: BSD 3 clause

import numpy as np

from libc.math cimport INFINITY, ceil, exp, fabs, floor, pow, sqrt
from libc.stdlib cimport free, malloc, qsort
from libc.string cimport memcpy, memset

from ...utils cimport TSArray, _stats


cdef extern from "catch22.h":

    cdef double histogram_mode(
        const double *x, int length, int *hist_counts, double *bin_edges, int n_bins
    ) nogil

    cdef double histogram_ami_even(
        const double *x, int length, int tau, int n_bins
    ) nogil

    cdef double transition_matrix_ac_sumdiagcov(
        const double *x, double *ac, int length, int n_groups
    ) nogil

    cdef double periodicity_wang_th0_01(const double *x, int length) nogil

    cdef double embed2_dist_tau_d_expfit_meandiff(
        const double *x, double *ac, int length
    ) nogil

    cdef double auto_mutual_info_stats_gaussian_fmmi(
        const double *x, int length, int tau
    ) nogil

    cdef double outlier_include_np_mdrmd(
        const double *x, int length, int sign, double inc
    ) nogil

    cdef double summaries_welch_rect(
        const double *x, int length, int what, double *S, double *f, int n_welch
    ) nogil

    cdef double motif_three_quantile_hh(const double *x, int length) nogil

    cdef double fluct_anal_2_50_1_logi_prop_r1(
        const double *y, int length, int lag, int how
    ) nogil


cdef double transition_matrix_3ac_sumdiagcov(
    const double *x, double *ac, Py_ssize_t length,
) noexcept nogil:
    return transition_matrix_ac_sumdiagcov(x, ac, length, 3)


cdef double histogram_ami_even_2_5(const double *x, Py_ssize_t length) noexcept nogil:
    return histogram_ami_even(x, length, 2, 5)


cdef double histogram_mode10(
    const double *x, Py_ssize_t length, int *bin_count, double *bin_edges
) noexcept nogil:
    return histogram_mode(x, length, bin_count, bin_edges, 10)


cdef double histogram_mode5(
    const double *x, Py_ssize_t length, int *bin_count, double *bin_edges
) noexcept nogil:
    return histogram_mode(x, length, bin_count, bin_edges, 5)


cdef double f1ecac(const double *ac, Py_ssize_t n) noexcept nogil:
    if n <= 1:
        return 0.0

    cdef double threshold = 1.0 / exp(1.0)
    cdef Py_ssize_t i
    for i in range(n - 1):
        if (ac[i] - threshold) * (ac[i + 1] - threshold) < 0:
            return i + 1
    return n


cdef double first_min(double *ac, Py_ssize_t n) noexcept nogil:
    if n <= 2:
        return 0.0
    
    cdef Py_ssize_t i
    for i in range(1, n - 1):
        if ac[i] < ac[i - 1] and ac[i] < ac[i + 1]:
            return i
    return n


cdef double trev_1_num(const double *x, Py_ssize_t n) noexcept nogil:
    if n <= 1:
        return 0.0

    cdef Py_ssize_t i
    cdef double sum = 0
    for i in range(n - 1):
        sum += pow(x[i + 1] - x[i], 3.0)
    return sum / (n - 1)


cdef double local_mean_std(const double *x, Py_ssize_t n, Py_ssize_t lag) noexcept nogil:
    if n <= lag:
        return 0.0

    cdef _stats.IncStats inc_stats
    _stats.inc_stats_init(&inc_stats)
    cdef Py_ssize_t i, j
    cdef double lag_sum
    for i in range(n - lag):
        lag_sum = 0.0
        for j in range(lag):
            lag_sum += x[i + j]

        _stats.inc_stats_add(
            &inc_stats, 1.0, x[i + lag] - lag_sum / lag
        )

    return sqrt(_stats.inc_stats_variance(&inc_stats, True))


cdef double hrv_classic_pnn(const double *x, Py_ssize_t n, double pnn) noexcept nogil:
    if n <= 1:
        return 0.0
    
    cdef Py_ssize_t i
    cdef double value = 0

    for i in range(1, n):
        if fabs(x[i] - x[(i - 1)]) * 1000 > pnn:
            value += 1

    return value / (n - 1)


cdef double above_mean_stretch(const double *x, Py_ssize_t n) noexcept nogil:
    cdef double mean = _stats.mean(x, n)
    cdef double stretch = 0
    cdef double longest = 0
    cdef Py_ssize_t i

    for i in range(n):
        if x[i] - mean <= 0:
            if stretch >= longest:
                longest = stretch
            stretch = 1
        else:
            stretch += 1

    if stretch > longest:
        return stretch
    else:
        return longest


cdef double below_diff_stretch(const double *x, Py_ssize_t n) noexcept nogil:
    cdef double stretch = 0
    cdef double max_stretch = 0
    cdef double last_i = 0
    cdef Py_ssize_t i

    for i in range(n - 1):
        if x[i + 1] - x[i] >= 0 or i == n - 2:
            stretch = i - last_i
            if stretch > max_stretch:
                max_stretch = stretch
            last_i = i

    return max_stretch


cdef double local_mean_tauresrat(
    const double *x, double *ac, Py_ssize_t n, Py_ssize_t lag
) noexcept nogil:
    if n <= lag or lag == 0:
        return 0.0
    cdef:
        Py_ssize_t i
        Py_ssize_t j
        double lag_sum
        Py_ssize_t lag_out, x_out
        double *lag_ac
    
    lag_ac = <double*> malloc(sizeof(double) * n - lag)
    for i in range(n - lag):
        lag_sum = 0
        for j in range(lag):
            lag_sum += x[i + j]

        lag_ac[i] = x[i + lag] - lag_sum / lag

    lag_out = 0
    _stats.auto_correlation(lag_ac, n - lag, lag_ac)
    while lag_ac[lag_out] > 0 and lag_out < n - lag:
        lag_out += 1
    free(lag_ac)

    x_out = 0
    while ac[x_out] > 0 and x_out < n:
        x_out += 1

    if x_out == 0:
        x_out = 1
    return <double> lag_out / <double> x_out



# Authors: Isak Samsten
# License: BSD 3 clause

"""
Wildboar - a library for temporal machine learning.

Wildboar includes numerous temporal machine learning algorithms and seamlessly
integrates them with `scikit-learn <https://scikit-learn.org>`__.
"""

import warnings

import numpy as np
from sklearn.utils.deprecation import deprecated

from .utils.variable_len import EOS, EoS, eos, is_end_of_series
from .version import version as __version__

__all__ = [
    "__version__",
    "eos",
    "EoS",
    "EOS",
    "iseos",  # TODO(1.3): Remove
]


# TODO(1.3): Remove
@deprecated(
    "To be removed in 1.3. Use wildboar.utils.variable_len.is_end_of_series instead."
)
def iseos(x):
    """
    Boolean indicator if a value is the end-of-sequence value.

    Parameters
    ----------
    x : array-like
        The array.

    Returns
    -------
    ndarray
        Boolean indicator array.
    """
    isneginf = np.isneginf(x)
    if isneginf.any():
        # TODO(1.3)
        warnings.warn(
            "Using -np.inf as eos has been deprecated in 1.2 and support will "
            "be removed in 1.3. Use wildboar.utils.variable_len.is_end_of_series",
            DeprecationWarning,
        )
        return np.logical_or(isneginf, is_end_of_series(x))

    return is_end_of_series(x)



# Authors: Isak Samsten
# License: BSD 3 clause

import numpy as np
import pytest
from sklearn import clone
from sklearn.utils._param_validation import (
    InvalidParameterError,
    generate_invalid_param_val,
    generate_valid_param,
    make_constraint,
)
from sklearn.utils._testing import MinimalClassifier

from ..base import BaseEstimator, is_explainer

_DUMMY_X = np.zeros((10, 10))
_DUMMY_X.setflags(write=False)
_DUMMY_Y = np.zeros(10)
_DUMMY_Y[:5] = 1
_DUMMY_Y.setflags(write=False)


def assert_exhaustive_parameter_checks(estimator: BaseEstimator):
    """
    Assert that all parameter are checked.

    Parameters
    ----------
    estimator : BaseEstimator
       The estimator to check.

    Attributes
    ----------
    test : bool
        Ok.
    """
    assert hasattr(estimator.__class__, "_parameter_constraints")
    assert (
        estimator.get_params(deep=False).keys()
        == estimator.__class__._parameter_constraints.keys()
    )


def assert_parameter_checks(estimator: BaseEstimator, skip=None):
    """
    Assert that all parameter checks are correct.

    Extended.

    Parameters
    ----------
    estimator : BaseEstimator
        The estimator.
    skip : list, optional
        The parameter constraints to skip.
    """
    assert hasattr(estimator.__class__, "_parameter_constraints")
    if is_explainer(estimator):
        clf = MinimalClassifier()
        clf.fit(_DUMMY_X, _DUMMY_Y)
        clf.n_features_in_ = _DUMMY_X.shape[1]

    for param, constraints in estimator.__class__._parameter_constraints.items():
        if skip is not None and param in skip:
            continue

        for constraint in (make_constraint(constraint) for constraint in constraints):
            try:
                invalid_value = generate_invalid_param_val(constraint)
            except NotImplementedError:
                invalid_value = None

            if invalid_value is not None:
                estimator_ = clone(estimator)
                estimator_.set_params(**{param: invalid_value})
                with pytest.raises(InvalidParameterError):
                    if is_explainer(estimator):
                        try:
                            estimator_._validate_params()
                        except Exception as e:
                            if isinstance(e, InvalidParameterError):
                                raise e
                    else:
                        try:
                            estimator_._validate_params()
                        except Exception as e:
                            if isinstance(e, InvalidParameterError):
                                raise e

            try:
                valid_value = generate_valid_param(constraint)
            except Exception:
                continue

            estimator_ = clone(estimator)
            estimator_.set_params(**{param: valid_value})
            if is_explainer(estimator):
                try:
                    estimator_._validate_params()
                except Exception as e:
                    if isinstance(e, InvalidParameterError):
                        raise e

            else:
                try:
                    estimator_._validate_params()
                except Exception as e:
                    if isinstance(e, InvalidParameterError):
                        raise e



import math
from functools import partial
from numbers import Integral, Real

import numpy as np
from sklearn.base import TransformerMixin
from sklearn.utils._param_validation import Interval

from ..base import BaseEstimator
from ..distance import matrix_profile


class MatrixProfileTransform(TransformerMixin, BaseEstimator):
    """
    Matrix profile transform.

    Transform each time series in a dataset to its MatrixProfile similarity
    self-join.

    Parameters
    ----------
    window : int or float, optional
        The subsequence size, by default 0.1.

        - if float, a fraction of n_timestep.
        - if int, the exact subsequence size.
    exclude : int or float, optional
        The size of the exclusion zone. The default exclusion zone is 0.2.

        - if float, expressed as a fraction of the windows size.
        - if int, exact size (0 < exclude).
    n_jobs : int, optional
        The number of jobs to use when computing the profile.

    Examples
    --------
    >>> from wildboar.datasets import load_two_lead_ecg()
    >>> from wildboar.transform import MatrixProfileTransform
    >>> x, y = load_two_lead_ecg()
    >>> t = MatrixProfileTransform()
    >>> t.fit_transform(x)
    """

    _parameter_constraints: dict = {
        "window": [
            Interval(Integral, 1, None, closed="left"),
            Interval(Real, 0, 1, closed="right"),
        ],
        "exclude": [
            None,
            Interval(Integral, 0, None, closed="left"),
            Interval(Real, 0, 1, closed="both"),
        ],
        "n_jobs": [None, Integral],
    }

    def __init__(self, window=0.1, exclude=None, n_jobs=None):
        self.window = window
        self.n_jobs = n_jobs
        self.exclude = exclude

    def fit(self, x, y=None):
        """
        Fit the matrix profile.

        Sets the expected input dimensions.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timesteps) \
        or (n_samples, n_dims, n_timesteps)
            The samples.
        y : ignored
            The optional labels.

        Returns
        -------
        self 
            A fitted instance.
        """
        self._validate_params()
        self._validate_data(x, dtype=float, allow_3d=True)

        if self.window > self.n_timesteps_in_:
            raise ValueError(
                f"The window parameter of {type(self).__name__} must be <= X.shape[-1]"
            )
        return self

    def transform(self, x):
        """
        Transform the samples to their MatrixProfile self-join.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timesteps) \
        or (n_samples, n_dims, n_timesteps)
            The samples.

        Returns
        -------
        ndarray of shape (n_samples, n_timestep) or (n_samples, n_dims, n_timesteps)
            The matrix matrix profile of each sample.
        """
        x = self._validate_data(x, reset=False, allow_3d=True, dtype=float)

        if isinstance(self.window, Integral):
            window = self.window
        else:
            window = math.ceil(self.window * self.n_timesteps_in_)

        func = partial(
            matrix_profile, window=window, exclude=self.exclude, n_jobs=self.n_jobs
        )
        if self.n_dims_in_ > 1:
            profile_size = self.n_timesteps_in_ - window + 1
            mp = np.empty((x.shape[0], self.n_dims_in_, profile_size), dtype=float)

            for i in range(self.n_dims_in_):
                mp[:, i, :] = func(x, dim=i)

            return mp
        else:
            return func(x)

    def _more_tags(self):
        return {"X_types": ["2darray", "3darray"]}



# Authors: Isak Samsten
# License: BSD 3 clause

import operator
import re

import numpy as np

from ..utils.validation import check_option

__all__ = [
    "make_dict_filter",
    "make_str_filter",
    "make_list_filter",
    "make_filter",
]


def __n_dims_filter(verb, dataset, x, y):
    n_dims = x.shape[1] if x.ndim > 2 else 1
    return __verb_compare(verb, n_dims)


def __casting_op(op, lhs, rhs):
    lhs_class = lhs.__class__
    return op(lhs, lhs_class(rhs))


def __verb_compare(verb, value):
    match = re.match(__VERB_PATTERN, verb)
    if match:
        return _OPERATORS[match.group(1)](value, match.group(2))
    else:
        raise ValueError(
            "Invalid comparision %s, must match %s" % (verb, __VERB_PATTERN)
        )


def __new_composite_filter(filters):
    def f(dataset, x, y):
        for filter in filters:
            if not filter(dataset, x, y):
                return False
        return True

    return f


def make_filter(filter):
    """
    Create a new filter.

    Parameters
    ----------
    filter : str, list or dict
        The filter.

    Returns
    -------
    callable
        The filter function.
    """
    if isinstance(filter, list):
        return make_list_filter(filter)
    elif isinstance(filter, dict):
        return make_dict_filter(filter)
    elif isinstance(filter, str):
        return make_str_filter(filter)
    else:
        raise TypeError(
            "filter must be list, dict or str, not %s" % type(filter).__qualname__
        )


def make_list_filter(filter):
    """
    Make a new filter based on a list of filter strings.

    Parameters
    ----------
    filter : list
        A list of filter strings.

    Returns
    -------
    callable
        The filter function.
    """
    filters = []
    for str_filter in filter:
        filters.append(make_str_filter(str_filter))

    return __new_composite_filter(filters)


def make_dict_filter(filter):
    """
    Make a new filter.

    Parameters
    ----------
    filter : dict
        The dict of `[subject] -> [op][verb]`.

    Returns
    -------
    callable
        The filter function.
    """
    filters = []
    for subject, verb in filter.items():
        filters.append(
            lambda dataset, x, y: check_option(_SUBJECTS, subject, "subject")(
                verb, dataset, x, y
            )
        )

    return __new_composite_filter(filters)


def make_str_filter(filter):
    """
    Make a new filter.

    Parameters
    ----------
    filter : str
        The filter string `[subject][op][verb]`.

    Returns
    -------
    callable
        The filter function.
    """
    match = re.match(__SUBJECT_VERB_PATTERN, filter)
    if match:
        subject = match.group(1)
        verb = match.group(2)
    else:
        raise ValueError(
            "Invalid filter %s, must match %s" % (filter, __SUBJECT_VERB_PATTERN)
        )

    def f(dataset, x, y):
        return check_option(_SUBJECTS, subject, "subject")(verb, dataset, x, y)

    return f


__VERB_PATTERN = re.compile(r"^(<|<=|>=|>|=|=~)\s*(\w+)$")
__SUBJECT_VERB_PATTERN = re.compile(r"^(\w+)\s*((?:<|<=|>=|>|=|=~)\s*\w+)$")


_SUBJECTS = {
    "dataset": lambda verb, dataset, x, y: __verb_compare(verb, dataset),
    "n_samples": lambda verb, dataset, x, y: __verb_compare(verb, x.shape[0]),
    "n_timestep": lambda verb, dataset, x, y: __verb_compare(verb, x.shape[-1]),
    "n_dims": __n_dims_filter,
    "n_labels": lambda verb, dataset, x, y: __verb_compare(verb, np.unique(y).shape[0]),
}

_OPERATORS = {
    "=~": lambda x, y: y in x,
    "=": lambda x, y: __casting_op(operator.eq, x, y),
    "<=": lambda x, y: __casting_op(operator.le, x, y),
    ">=": lambda x, y: __casting_op(operator.ge, x, y),
    "<": lambda x, y: __casting_op(operator.lt, x, y),
    ">": lambda x, y: __casting_op(operator.gt, x, y),
}



import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from wildboar.base import is_explainer
from wildboar.datasets import load_two_lead_ecg
from wildboar.explain import AmplitudeImportance, IntervalImportance, ShapeletImportance
from wildboar.tree import ShapeletTreeClassifier
from wildboar.utils._testing import (
    assert_exhaustive_parameter_checks,
    assert_parameter_checks,
)


@pytest.mark.parametrize(
    "explainer",
    [
        IntervalImportance(),
        AmplitudeImportance(),
        ShapeletImportance(),
    ],
)
def test_parameter_constrains(explainer):
    assert_exhaustive_parameter_checks(explainer)
    assert_parameter_checks(explainer)


@pytest.mark.parametrize(
    "explainer",
    [
        IntervalImportance(),
        AmplitudeImportance(),
        ShapeletImportance(),
    ],
)
def test_is_explainer(explainer):
    assert is_explainer(explainer)


def test_interval_importance():
    x_train, x_test, y_train, y_test = load_two_lead_ecg(False)
    clf = ShapeletTreeClassifier(n_shapelets=1000, random_state=123)
    clf.fit(x_train, y_train)

    imp = IntervalImportance(scoring="accuracy", random_state=123)
    imp.fit(clf, x_test, y_test)

    actual_scores = imp.importances_.mean
    expected_scores = np.array(
        [
            0.0064969,
            0.0149254,
            0.01036,
            0.2363477,
            0.2361721,
            0.2186128,
            0.0612818,
            0.0273924,
            0.0231782,
            0.0061457,
        ]
    )

    assert_almost_equal(actual_scores, expected_scores)


def test_amplitude_importance():
    x_train, x_test, y_train, y_test = load_two_lead_ecg(False)
    clf = ShapeletTreeClassifier(n_shapelets=1000, random_state=123)
    clf.fit(x_train, y_train)

    imp = AmplitudeImportance(scoring="accuracy", random_state=123)
    imp.fit(clf, x_test, y_test)

    actual_scores = imp.importances_.mean
    expected_scores = np.array([0.2712906, 0.191396, 0.1018437, 0.0570676])

    assert_almost_equal(actual_scores, expected_scores)


def test_shapelet_importance():
    x_train, x_test, y_train, y_test = load_two_lead_ecg(False)
    clf = ShapeletTreeClassifier(n_shapelets=1000, random_state=123)
    clf.fit(x_train, y_train)

    imp = ShapeletImportance(scoring="accuracy", random_state=123)
    imp.fit(clf, x_test, y_test)

    actual_scores = imp.importances_.mean
    expected_scores = np.array(
        [
            0.3028973,
            0.2431958,
            0.1931519,
            0.2835821,
            0.2756804,
            0.260755,
            0.2186128,
            0.2941176,
            0.1729587,
            -0.0087796,
        ]
    )

    assert_almost_equal(actual_scores, expected_scores)



# Authors: Isak Samsten
# License: BSD 3 clause

import numbers

import numpy as np
from sklearn.utils.multiclass import type_of_target
from sklearn.utils.validation import (
    _check_estimator_name,
    _check_y,
    check_consistent_length,
    warnings,
)
from sklearn.utils.validation import check_array as sklearn_check_array

from .variable_len import is_end_of_series, is_variable_length


def check_classification_targets(y):
    """
    Ensure that a classification target is either binary of multiclass.

    Parameters
    ----------
    y : array-like
        The target.

    Raises
    ------
    ValueError
        If the type of target is not binary of multiclass
    """
    y_type = type_of_target(y, input_name="y")
    if y_type not in ["binary", "multiclass"]:
        raise ValueError(
            f"Unknown label type: {y_type}. Maybe you are trying to fit a "
            "classifier, which expects discrete classes on a "
            "regression target with continuous values."
        )


# noqa: H0002
def check_type(x, name, target_type, required=True):
    """
    Check that the type of x is of a target type.

    Parameters
    ----------
    x : object
        The object to check.
    name : str
        The name of the parameter.
    target_type : type or tuple
        The required type(s) of x.
    required : bool, optional
        If required=False, None is an allowed.
    """

    def type_name(t):
        module = t.__module__
        qualname = t.__qualname__
        if module == "builtins":
            return qualname
        elif t == numbers.Real:
            return "float"
        elif t == numbers.Integral:
            return "int"
        return f"{module}.{qualname}"

    if isinstance(target_type, tuple):
        types_str = ", ".join(type_name(t) for t in target_type)
        target_type_str = f"{{{types_str}}}"
    else:
        target_type_str = type_name(target_type)

    if x is None and required:
        raise TypeError(f"{name} must be an instance of {target_type_str}, not None")
    if x is None and not required:
        return

    if not isinstance(x, target_type):
        raise TypeError(
            f"{name} must be an instance of {target_type_str}, not"
            f" {type(x).__qualname__}."
        )


# noqa: H0002
def check_option(options, key, name):
    """
    Check that `key` is a valid option.

    Parameters
    ----------
    options : dict
        The dictionary of option value pairs.
    key : str
        The option key.
    name : str
        The name of the parameter.

    Returns
    -------
    option_value
        The value of `key` in `option`.

    Raises
    ------
    ValueError
       If `key` is not a valid option.
    """
    if key in options:
        return options[key]
    else:
        keys = ["'%s'" % key for key in sorted(list(options.keys()))]
        if len(keys) == 1:
            msg = f"{name} must be {keys[0]}, got {key}"
        else:
            msg = f"{name} must be {', '.join(keys[:-1])} or {keys[-1]}, got {key}"

        raise ValueError(msg)


def _num_timesteps_dim(dim):
    type_ = type(dim)
    if type_.__module__ == "builtins":
        type_name = type_.__qualname__
    else:
        type_name = f"{type_.__module__}.{type_.__qualname__}"
    message = (
        f"Unable to find the number of timesteps from dimension of type {type_name}"
    )
    if not hasattr(dim, "__len__") and not hasattr(dim, "shape"):
        if not hasattr(dim, "__array__"):
            raise TypeError(message)
        dim = np.asarray(dim)

    if hasattr(dim, "shape"):
        if not hasattr(dim, "__len__") or len(dim.shape) != 1:
            message += f" with shape {dim.shape}"
            raise TypeError(message)

        return dim.shape[0]

    if len(dim) > 0:
        if (
            hasattr(dim, "__len__")
            or hasattr(dim, "shape")
            or hasattr(dim, "__array__")
        ):
            raise TypeError(message)

    return len(dim)


def _num_timesteps(X):
    """
    Return the number of timesteps and dimensions of `X`.

    Parameters
    ----------
    X : object
        The object to guess the number of timesteps and dimensions

    Returns
    -------
    n_timesteps : int
        The number of timesteps
    n_dims : int
        The number of dimensions
    """
    type_ = type(X)
    if type_.__module__ == "builtins":
        type_name = type_.__qualname__
    else:
        type_name = f"{type_.__module__}.{type_.__qualname__}"
    message = f"Unable to find the number of timesteps from X of type {type_name}"
    if not hasattr(X, "__len__") and not hasattr(X, "shape"):
        if not hasattr(X, "__array__"):
            raise TypeError(message)
        # Only convert X to a numpy array if there is no cheaper, heuristic
        # option.
        X = np.asarray(X)

    if hasattr(X, "shape"):
        if not hasattr(X.shape, "__len__") or len(X.shape) <= 1 or len(X.shape) > 3:
            message += f" with shape {X.shape}"
            raise TypeError(message)
        return X.shape[-1], X.shape[1] if len(X.shape) == 3 else 1

    first_sample = X[0]

    # Do not consider an array-like of strings or dicts to be a 2D array
    if isinstance(first_sample, (str, bytes, dict)):
        message += f" where the samples are of type {type(first_sample).__qualname__}"
        raise TypeError(message)

    try:
        possibly_first_dim = first_sample[0]
        if hasattr(possibly_first_dim, "__len__") or hasattr(
            possibly_first_dim, "shape"
        ):
            return _num_timesteps_dim(possibly_first_dim), len(first_sample)

        # If X is a list of lists, for instance, we assume that all nested
        # lists have the same length without checking or converting to
        # a numpy array to keep this function call as cheap as possible.
        return len(first_sample), 1
    except Exception as err:
        raise TypeError(message) from err


def check_X_y(  # noqa: PLR0913, N802
    x,
    y,
    *,
    dtype=float,
    order="C",
    copy=False,
    ensure_2d=True,
    ensure_ts_array=False,
    allow_3d=False,
    allow_nd=False,
    force_all_finite=True,
    multi_output=False,
    ensure_min_samples=1,
    ensure_min_timesteps=1,
    ensure_min_dims=1,
    allow_eos=False,
    y_numeric=False,
    y_contiguous=True,
    estimator=None,
):
    """
    Check both X and y.

    Parameters
    ----------
    x : array-like
        The samples.
    y : array-like
        The labels.
    dtype : dtype, optional
        The data type for `X`.
    order : {"C", "F"}, optional
        The order of data in memory.
    copy : bool, optional
        Force a copy of `X`.
    ensure_2d : bool, optional
        Ensure that the array is 2d, i.e., (n_samples, n_timesteps).
    ensure_ts_array : bool, optional
        Ensure that the array is a valid time series array.
    allow_3d : bool, optional
        Allow `X` to be 3d, i.e., (n_samples, n_dimensions, n_timesteps).
    allow_nd : bool, optional
        Allow `X` to have 2 or more dimensions.
    force_all_finite : bool, optional
        Require every value in `X` to be finite.
    multi_output : bool, optional
        Allow `y` to be a multi output array.
    ensure_min_samples : int, optional
        Require `X` to have at least this many samples.
    ensure_min_timesteps : int, optional
        Require `X` to have at least this many timesteps.
    ensure_min_dims : int, optional
        Require `X` to have at least this many dimensions.
    allow_eos : bool, optional
        Allow `X` to be of variale length.
    y_numeric : bool, optional
        Ensure that `y` is numeric with dtype `float`.
    y_contiguous : bool, optional
        Ensure that `y` is memory contiguous.
    estimator : object, optional
        An estimator object for error reporting.

    Returns
    -------
    X : ndarray
        The validated array `X`.
    y : ndarray
        The validated array `y`.
    """
    if y is None:
        if estimator is None:
            estimator_name = "estimator"
        else:
            estimator_name = _check_estimator_name(estimator)
        raise ValueError(
            f"{estimator_name} requires y to be passed, but the target y is None"
        )
    x = check_array(
        x,
        allow_3d=allow_3d,
        allow_nd=allow_nd,
        ensure_min_samples=ensure_min_samples,
        ensure_min_dims=ensure_min_dims,
        ensure_min_timesteps=ensure_min_timesteps,
        force_all_finite=force_all_finite,
        allow_eos=allow_eos,
        order=order,
        copy=copy,
        ensure_2d=ensure_2d,
        ensure_ts_array=ensure_ts_array,
        dtype=dtype,
    )

    if y_numeric:
        y_dtype = np.float64
    else:
        y_dtype = None

    if y_contiguous:
        y = np.ascontiguousarray(y, dtype=y_dtype)

    y = _check_y(y, multi_output=multi_output, y_numeric=y_numeric, estimator=estimator)
    check_consistent_length(x, y)
    return x, y


def check_array(  # noqa: PLR0913, PLR0912
    array,
    *,
    dtype="numeric",
    order="C",
    copy=False,
    ravel_1d=False,
    ensure_2d=True,
    ensure_ts_array=False,
    allow_3d=False,
    allow_nd=False,
    allow_eos=False,
    force_all_finite=True,
    ensure_min_samples=1,
    ensure_min_timesteps=1,
    ensure_min_dims=1,
    estimator=None,
    input_name="",
):
    """
    Input validation on time-series.

    Delegate array validation to scikit-learn
    :func:`sklearn.utils.validation.check_array` with wildboar defaults and
    conventions.

    - we optionally allow end-of-sequence identifiers
    - by default we convert arrays to c-order
    - we optionally specifically allow for 3d-arrays
    - we never allow for sparse arrays

    By default, the input is checked to be a non-empty 2D array in c-order containing
    only finite values, with at least 1 sample, 1 timestep and 1 dimension. If the dtype
    of the array is object, attempt converting to float, raising on failure.

    Parameters
    ----------
    array : object
        Input object to check / convert.
    dtype : 'numeric', type, list of type or None, optional
        Data type of result. If None, the dtype of the input is preserved.
        If "numeric", dtype is preserved unless array.dtype is object.
        If dtype is a list of types, conversion on the first type is only
        performed if the dtype of the input is not in the list.
    order : {'F', 'C', 'T'} or None, optional
        Whether an array will be forced to be fortran or c-style.
        When order is None, then if copy=False, nothing is ensured
        about the memory layout of the output array; otherwise (copy=True)
        the memory layout of the returned array is kept as close as possible
        to the original array.
    copy : bool, optional
        Whether a forced copy will be triggered. If copy=False, a copy might
        be triggered by a conversion.
    ravel_1d : bool, optional
        Whether to ravel 1d arrays or column vectors, it the array is neither an
        error is raised.
    ensure_2d : bool, optional
        Whether to raise a value error if array is not 2D.
    allow_3d : bool, optional
        Whether to allow array.ndim == 3.
    allow_nd : bool, optional
        Whether to allow array.ndim > 2.
    allow_eos : bool, optional
        Whether to raise an error on `wildboar.utils.variable_len.eos` in the
        array.
    force_all_finite : bool or 'allow-nan', default=True
        Whether to raise an error on np.inf, np.nan, pd.NA in array. The
        possibilities are:

        - True: Force all values of array to be finite.
        - False: accepts np.inf, np.nan, pd.NA in array.
        - 'allow-nan': accepts only np.nan and pd.NA values in array. Values
          cannot be infinite.
    ensure_min_samples : int, optional
        Make sure that the array has a minimum number of samples in its first
        axis (rows for a 2D array). Setting to 0 disables this check.
    ensure_min_timesteps : int, optional
        Make sure that the 2D array has some minimum number of timesteps
        (columns). The default value of 1 rejects empty datasets.
        This check is only enforced when the input data has effectively 2
        dimensions or is originally 1D and ``ensure_2d`` is True. Setting to 0
        disables this check.
    ensure_min_dims : int, optional
        Make sure that the array has a minimum number of dimensions. Setting to 0
        disables this check.
    estimator : str or estimator instance, default=None
        If passed, include the name of the estimator in warning messages.
    input_name : str, default=""
        The data name used to construct the error message.

    Returns
    -------
    object
        The converted and validated array.
    """
    check_params = dict(
        accept_sparse=False,
        accept_large_sparse=False,
        dtype=dtype,
        order=None if ensure_ts_array else order,
        copy=copy,
        force_all_finite=False,
        ensure_2d=ensure_2d,
        allow_nd=allow_3d or allow_nd,
        ensure_min_samples=ensure_min_samples,
        ensure_min_features=0,
        estimator=estimator,
        input_name=input_name,
    )
    if force_all_finite not in (True, False, "allow-nan"):
        raise ValueError(
            'force_all_finite should be a bool or "allow-nan".'
            f"Got {force_all_finite} instead"
        )

    array = sklearn_check_array(array, **check_params)
    estimator_name = _check_estimator_name(estimator)
    if not (allow_nd or allow_3d) and array.ndim >= 3:
        raise ValueError(
            "Found array with dim %d. %s expected <= 2." % (array.ndim, estimator_name)
        )

    if not allow_nd and allow_3d and array.ndim >= 4:
        raise ValueError(
            "Found array with dim %d. %s expected <= 3." % (array.ndim, estimator_name)
        )

    if ravel_1d:
        if array.ndim == 1:
            return array.ravel(order=order)
        elif array.ndim == 2 and array.shape[1] == 1:
            return array.ravel(order=order)
        else:
            raise ValueError(
                "Found array with dim %d.%s expect 1 dim or column vector"
                % (array.ndim, estimator_name)
            )

    context = " by %s" % estimator_name if estimator is not None else ""
    if ensure_min_dims > 0 and (array.ndim == 2 or array.ndim == 3):  # noqa: PLR1714
        if array.ndim == 3:
            n_dims = array.shape[1]
        else:
            n_dims = 1

        if n_dims < ensure_min_dims:
            raise ValueError(
                "Found array with %d dimension(s) (shape=%s) while a"
                " minimum of %d is required%s."
                % (n_dims, array.shape, ensure_min_dims, context)
            )

    if ensure_min_timesteps > 0 and (array.ndim == 2 or array.ndim == 3):  # noqa: PLR1714
        n_timesteps = array.shape[-1]
        if n_timesteps < ensure_min_timesteps:
            # TODO: ignore sklearn tests for this error message and create our own
            #       estimator check
            raise ValueError(
                "Found array with %d feature(s) (shape=%s) while"
                " a minimum of %d is required%s."
                % (n_timesteps, array.shape, ensure_min_timesteps, context)
            )

    if np.issubdtype(array.dtype, np.double):
        padded_input_name = input_name + " " if input_name else ""
        if not allow_eos and is_variable_length(array):
            raise ValueError(
                f"Input {padded_input_name}expected time series of equal length."
            )

        if force_all_finite is True:
            if allow_eos:
                anynan = np.logical_and(np.isnan(array), ~is_end_of_series(array)).any()
            else:
                anynan = np.isnan(array).any()

            if anynan:
                raise ValueError(f"Input {padded_input_name}contains NaN.")

        if force_all_finite and np.isinf(array).any():
            if allow_eos and not np.isposinf(array).any():
                # TODO(1.3)
                warnings.warn(
                    "Using -np.inf as eos has been deprecated in 1.3 and support will "
                    "be removed in 1.3. ",
                    DeprecationWarning,
                )
            else:
                raise ValueError(f"Input {padded_input_name}contains infinity.")

    return _check_ts_array(array) if ensure_ts_array else array


def _check_ts_array(array):
    """
    Ensure a time-series array.

    Force the array to (1) be 3D (n_samples, n_dims, n_timesteps) and (2)
    have the final dimension contiguous in memory, and (3) have dtype=float.

    Parameters
    ----------
    array : ndarray
        The array

    Returns
    -------
    tsarray : TSArray
        A time series array
    """
    if array.ndim == 1:
        array = array.reshape(1, 1, array.shape[0])
    elif array.ndim == 2:
        array = array.reshape(array.shape[0], 1, array.shape[1])

    last_stride = array.strides[2] // array.itemsize
    if last_stride != 1:
        array = np.ascontiguousarray(array)

    return array.astype(float, copy=False)



# Authors: Isak Samsten
# License: BSD 3 clause

import warnings
from functools import partial
from unittest.case import SkipTest

import numpy as np
from sklearn.base import clone, is_classifier, is_regressor
from sklearn.exceptions import SkipTestWarning
from sklearn.utils._testing import (
    assert_allclose_dense_sparse,
    ignore_warnings,
    raises,
    set_random_state,
)
from sklearn.utils.estimator_checks import (
    _enforce_estimator_tags_y,
    _maybe_skip,
    _safe_tags,
)
from sklearn.utils.estimator_checks import (
    _yield_all_checks as _yield_all_checks_sklearn,
)
from sklearn.utils.validation import has_fit_parameter

from ..base import is_counterfactual, is_explainer


def _yield_classifier_test(estimator):
    if has_fit_parameter(estimator, "sample_weight"):
        yield partial(_check_sample_weights_invariance_samples_order, kind="ones")
        yield partial(_check_sample_weights_invariance_samples_order, kind="zeros")

    # if not _safe_tags(estimator).get("poor_score", False):
    #     yield check_decent_score


def _yield_regressor_checks(estimator):
    pass


def _yield_transform_checks(estimator):
    pass


def _yield_explainer_checks(estimator):
    pass


def _yield_all_checks(estimator):
    if is_classifier(estimator):
        for check in _yield_classifier_test(estimator):
            yield check

    if is_regressor(estimator):
        pass
    #     for check in _yield_regressor_checks(estimator):
    #         yield check

    if hasattr(estimator, "transform"):
        pass
    #     for check in _yield_transform_checks(estimator):
    #         yield check

    if is_explainer(estimator):
        pass
    #     for check in _yield_explainer_checks(estimator):
    #         yield check

    if is_counterfactual(estimator):
        pass

    if not is_explainer(estimator):
        if hasattr(estimator, "estimator_params"):
            yield _check_consistent_estimator_params

        if "3darray" in _safe_tags(estimator, "X_types"):
            yield _check_force_n_dims_raises
            yield _check_force_n_dims


# noqa: H0002
def check_estimator(estimator, generate_only=False, ignore=None, skip_scikit=False):
    """
    Check if estimator adheres to scikit-learn (and wildboar) conventions.

    This method delegates to `check_estimator` in scikit-learn but monkey-patches
    the estimator with tags to skip some tests related to performance.

    We also add a new parameter to allow for silently ignoring some scikit-learn
    tests.

    Finally, we also add a suite of specific wildboar-tests tailored towards
    time series classifiers, regressors and transformers.

    Parameters
    ----------
    estimator : estimator object
        Estimator instance to check.
    generate_only : bool, default=False
        When `False`, checks are evaluated when `check_estimator` is called.
        When `True`, `check_estimator` returns a generator that yields
        (estimator, check) tuples. The check is run by calling
        `check(estimator)`.
    ignore : list, optional
        Ignore the checks in the list.
    skip_scikit : bool, optional
        Skip all scikit-learn tests.

    Returns
    -------
    generator
        Generator that yields (estimator, check) tuples. Returned when
        `generate_only=True`.
    """

    if ignore is None:
        ignore = []

    name = type(estimator).__name__

    def checks_generator():
        if not skip_scikit:
            if estimator is not None:
                if hasattr(estimator, "_more_tags"):
                    _more_tags = estimator._more_tags().copy()
                    _more_tags.update({"poor_score": True})

                old_more_tags = estimator.__class__._more_tags

                def _new_more_tags_skip_low_score(self):
                    return _more_tags

                # Monkey-patch the estimator to always have the tag "poor_score" set
                # to true. We have specific tests in wildboar for testing performance.
                setattr(
                    estimator.__class__, "_more_tags", _new_more_tags_skip_low_score
                )

            for check in _yield_all_checks_sklearn(estimator):
                check_name = (
                    check.func.__name__
                    if isinstance(check, partial)
                    else check.__name__
                )

                # Silently ignore any scikit-learn tess in the ignore list
                if check_name not in ignore:
                    yield estimator, partial(_maybe_skip(estimator, check), name)

            if estimator is not None:
                # Reset the old _more_tags() method
                setattr(estimator.__class__, "_more_tags", old_more_tags)

        for check in _yield_all_checks(estimator):
            yield estimator, partial(_maybe_skip(estimator, check), name)

    if generate_only:
        return checks_generator()

    for estimator, check in checks_generator():
        try:
            check(estimator)
        except SkipTest as exception:
            warnings.warn(str(exception), SkipTestWarning)


@ignore_warnings(category=FutureWarning)
def _check_sample_weights_invariance_samples_order(name, estimator_orig, kind="ones"):
    # For kind="ones" check that the estimators yield same results for
    # unit weights and no weights
    # For kind="zeros" check that setting sample_weight to 0 is equivalent
    # to removing corresponding samples.
    estimator1 = clone(estimator_orig)
    estimator2 = clone(estimator_orig)
    if hasattr(estimator_orig, "bootstrap"):
        estimator1.bootstrap = False
        estimator2.bootstrap = False

    set_random_state(estimator1, random_state=0)
    set_random_state(estimator2, random_state=0)

    X1, y1 = _dummy_dataset()

    if kind == "ones":
        X2 = X1
        y2 = y1
        sw2 = np.ones(shape=len(y1))
        err_msg = (
            f"For {name} sample_weight=None is not equivalent to sample_weight=ones"
        )
    elif kind == "zeros":
        # Construct a dataset that is very different to (X, y) if weights
        # are disregarded, but identical to (X, y) given weights.
        X2 = np.vstack([X1, X1 + 1])
        y2 = np.hstack([y1, 3 - y1])
        sw2 = np.ones(shape=len(y1) * 2)
        sw2[len(y1) :] = 0

        # NOTE: We disable shuffling to allow support estimators that sample
        #
        # X2, y2, sw2 = shuffle(X2, y2, sw2, random_state=0)

        err_msg = (
            f"For {name}, a zero sample_weight is not equivalent to removing the sample"
        )
    else:  # pragma: no cover
        raise ValueError

    y1 = _enforce_estimator_tags_y(estimator1, y1)
    y2 = _enforce_estimator_tags_y(estimator2, y2)

    estimator1.fit(X1, y=y1, sample_weight=None)
    estimator2.fit(X2, y=y2, sample_weight=sw2)

    for method in ["predict", "predict_proba", "decision_function", "transform"]:
        if hasattr(estimator_orig, method):
            X_pred1 = getattr(estimator1, method)(X1)
            X_pred2 = getattr(estimator2, method)(X1)
            assert_allclose_dense_sparse(X_pred1, X_pred2, err_msg=err_msg)


def _dummy_dataset():
    X1 = np.array(
        [
            [1, 3],
            [1, 3],
            [1, 3],
            [1, 3],
            [2, 1],
            [2, 1],
            [2, 1],
            [2, 1],
            [3, 3],
            [3, 3],
            [3, 3],
            [3, 3],
            [4, 1],
            [4, 1],
            [4, 1],
            [4, 1],
        ],
        dtype=np.float64,
    )
    y1 = np.array([1, 1, 1, 1, 2, 2, 2, 2, 1, 1, 1, 1, 2, 2, 2, 2], dtype=int)
    return X1, y1


def _dataset_shape(*shape):
    return np.arange(np.prod(shape)).reshape(*shape), np.zeros(shape[0])


def _check_consistent_estimator_params(name, estimator):
    x, y = _dummy_dataset()
    estimator = clone(estimator)
    estimator.fit(x, y)

    for tree in estimator.estimators_:
        for param in estimator.estimator_params:
            assert getattr(tree, param) == getattr(estimator, param), (
                f"For {name} estimator_params are not equivalent "
                "to the base estimators parameters"
            )


def _check_force_n_dims_raises(name, estimator):
    X1, y1 = _dataset_shape(10, 3, 10)
    X2, y2 = _dataset_shape(10, 10)

    estimator = clone(estimator)
    estimator._force_n_dims = 2
    match = "has _force_n_dims set to 2"
    err_msg = (
        f"If _force_n_dims is set the {name} must reject "
        "3d arrays with unsupported n_dims."
    )

    with raises(
        ValueError,
        match=match,
        err_msg=err_msg,
    ):
        estimator.fit(X1, y1)

    if has_fit_parameter(estimator, "check_input"):
        with raises(ValueError, match=match, err_msg=err_msg):
            estimator.fit(X1, y1, check_input=False)

    del estimator._force_n_dims
    estimator.fit(X2, y2)
    estimator._force_n_dims = 2
    for method in ["predict", "predict_proba", "decision_function", "transform"]:
        if hasattr(estimator, method):
            with raises(ValueError, match=match, err_msg=err_msg):
                getattr(estimator, method)(X1)


def _check_force_n_dims(name, estimator):
    X1, y1 = _dataset_shape(10, 3, 10)
    X2, y2 = _dataset_shape(10, 3 * 10)
    estimator = clone(estimator)
    estimator.fit(X1, y1)

    estimator._force_n_dims = 3
    for method in ["predict", "predict_proba", "decision_function", "transform"]:
        if hasattr(estimator, method):
            getattr(estimator, method)(X2)



# cython: language_level=3

# Authors: Isak Samsten
# License: BSD 3 clause

cdef double histogram_mode10(
        const double *x,
        Py_ssize_t length,
        int *bin_count,
        double *bin_edges
) noexcept nogil

cdef double histogram_mode5(
        const double *x,
        Py_ssize_t length,
        int *bin_count,
        double *bin_edges
) noexcept nogil


cdef double histogram_ami_even_2_5(const double *x, Py_ssize_t length) noexcept nogil

cdef double transition_matrix_3ac_sumdiagcov(const double *x, double *ac, Py_ssize_t n) noexcept nogil

cdef double f1ecac(const double *ac, Py_ssize_t n) noexcept nogil

cdef double first_min(double *ac, Py_ssize_t n) noexcept nogil

cdef double trev_1_num(const double *x, Py_ssize_t n) noexcept nogil

cdef double local_mean_std(const double *x, Py_ssize_t n, Py_ssize_t lag) noexcept nogil

cdef double hrv_classic_pnn(const double *x, Py_ssize_t n, double pnn) noexcept nogil

cdef double above_mean_stretch(const double *x, Py_ssize_t n) noexcept nogil

cdef double below_diff_stretch(const double *x, Py_ssize_t n) noexcept nogil

cdef double local_mean_tauresrat(const double *x, double *ac, Py_ssize_t n, Py_ssize_t lag) noexcept nogil

cdef double periodicity_wang_th0_01(const double *x, int length) noexcept nogil

cdef double embed2_dist_tau_d_expfit_meandiff(const double *x, double *ac, int length) noexcept nogil

cdef double auto_mutual_info_stats_gaussian_fmmi(const double *x, int length, int tau) nogil 

cdef double outlier_include_np_mdrmd(const double *x, int length, int sign, double inc) noexcept nogil

cdef double summaries_welch_rect(
    const double *x, int length, int what, double *S, double *f, int n_welch
) noexcept nogil

cdef double motif_three_quantile_hh(const double *x, int length) noexcept nogil

cdef double fluct_anal_2_50_1_logi_prop_r1(const double *x, int length, int lag, int how) noexcept nogil



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3
# cython: initializedcheck=False

# Authors: Isak Samsten
# License: BSD 3 clause
import numpy as np
from libc.math cimport log, sqrt
from libc.stdlib cimport free, malloc


cdef inline uint32_t DEFAULT_SEED = 1

cdef class RandomSampler:

    def __cinit__(self, Py_ssize_t upper, const double[::1] weights=None):
        self.weights = weights
        self.upper = upper
        if self.weights is not None:
            if upper != weights.shape[0]:
                raise ValueError("upper != len(weights)")

            vose_rand_init(&self.vr, len(weights))
            vose_rand_precompute(&self.vr, &weights[0])

    def __reduce__(self):
        weights = np.asarray(self.weights) if self.weights is not None else None
        return self.__class__, (self.upper, weights)

    def __dealloc__(self):
        if self.weights is not None:
            vose_rand_free(&self.vr)

    cdef Py_ssize_t rand_int(self, uint32_t *seed) noexcept nogil:
        if self.weights is not None:
            return vose_rand_int(&self.vr, seed)
        else:
            return rand_int(0, self.upper, seed)


# https://jugit.fz-juelich.de/mlz/ransampl/-/blob/master/lib/ransampl.c
cdef void vose_rand_init(VoseRand *vr, Py_ssize_t n) noexcept nogil:
    vr.prob = <double*> malloc(sizeof(double) * n)
    vr.alias = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * n)
    vr.n = n

cdef void vose_rand_free(VoseRand *vr) noexcept nogil:
    free(vr.prob)
    free(vr.alias)

cdef void vose_rand_precompute(VoseRand *vr, const double *p) noexcept nogil:
    cdef Py_ssize_t n = vr.n
    cdef Py_ssize_t i, a, g
    cdef double *P = <double*> malloc(sizeof(double) * n)
    cdef Py_ssize_t *S = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * n)
    cdef Py_ssize_t *L = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * n)

    cdef double s = 0
    for i in range(n):
        s += p[i]

    for i in range(n):
        P[i] = p[i] * n / s

    cdef Py_ssize_t nS = 0
    cdef Py_ssize_t nL = 0
    for i in range(n - 1, -1, -1):
        if P[i] < 1:
            S[nS] = i
            nS += 1
        else:
            L[nL] = i
            nL += 1

    while nS > 0 and nL > 0:
        nS -= 1
        a = S[nS]
        nL -= 1
        g = L[nL]
        vr.prob[a] = P[a]
        vr.alias[a] = g
        P[g] = P[g] + P[a] - 1
        if P[g] < 1:
            S[nS] = g
            nS += 1
        else:
            L[nL] = g
            nL += 1

    while nL > 0:
        nL -= 1
        vr.prob[L[nL]] = 1

    while nS > 0:
        nS -= 1
        vr.prob[S[nS]] = 1

    free(P)
    free(S)
    free(L)


cdef Py_ssize_t vose_rand_int(VoseRand *vr, uint32_t *seed) noexcept nogil:
    cdef double r1 = rand_uniform(0, 1, seed)
    cdef double r2 = rand_uniform(0, 1, seed)
    cdef Py_ssize_t i = <Py_ssize_t> (vr.n * r1)
    if r2 < vr.prob[i]:
        return i
    else:
        return vr.alias[i]

# https://github.com/scikit-learn/scikit-learn/blob/
#  433600e68fbb12e72d8c5e0707916f5603bb7057/sklearn/utils/_random.pxd#L26
# TODO(1.2):
# if (seed[0] == 0): seed[0] = DEFAULT_SEED
#
#    seed[0] ^= <uint32_t>(seed[0] << 13)
#    seed[0] ^= <uint32_t>(seed[0] >> 17)
#    seed[0] ^= <uint32_t>(seed[0] << 5)
#    return seed[0] % ((<uint32_t>RAND_R_MAX) + 1)
cdef inline uint32_t rand_r(uint32_t *seed) noexcept nogil:
    """Returns a pesudo-random number based on the seed."""
    seed[0] = seed[0] * 1103515245 + 12345
    return seed[0] % (<uint32_t> RAND_R_MAX + 1)


cdef int rand_int(int min_val, int max_val, uint32_t *seed) noexcept nogil:
    """Returns a pseudo-random number in the range [`min_val` `max_val`["""
    if min_val == max_val:
        return min_val
    else:
        return min_val + rand_r(seed) % (max_val - min_val)


cdef double rand_uniform(
    double low, double high, uint32_t *random_state
) noexcept nogil:
    """Generate a random double in the range [`low` `high`[."""
    return ((high - low) * <double> rand_r(random_state) / <double> RAND_R_MAX) + low


cdef double rand_normal(double mu, double sigma, uint32_t *random_state) noexcept nogil:
    cdef double x1, x2, w, _y1
    x1 = 2.0 * rand_uniform(0, 1, random_state) - 1.0
    x2 = 2.0 * rand_uniform(0, 1, random_state) - 1.0
    w = x1 * x1 + x2 * x2
    while w >= 1.0:
        x1 = 2.0 * rand_uniform(0, 1, random_state) - 1.0
        x2 = 2.0 * rand_uniform(0, 1, random_state) - 1.0
        w = x1 * x1 + x2 * x2

    w = sqrt((-2.0 * log(w)) / w)
    _y1 = x1 * w
    y2 = x2 * w
    return mu + _y1 * sigma


cdef void shuffle(Py_ssize_t *values, Py_ssize_t length, uint32_t *seed) noexcept nogil:
    cdef Py_ssize_t i, j
    for i in range(length - 1, 0, -1):
        j = rand_int(0, i, seed)
        values[i], values[j] = values[j], values[i]



# Authors: Isak Samsten
# License: BSD 3 clause

import numpy as np
import pytest

from wildboar.utils.validation import check_array
from wildboar.utils.variable_len import eos


def test_check_array_multivariate():
    x = np.arange(10 * 3 * 10).reshape(10, 3, 10)
    x_checked = check_array(x, allow_3d=True)
    assert x.dtype == x_checked.dtype

    with pytest.raises(ValueError):
        check_array(x, allow_3d=False)


def test_check_array_ensure_min_dims():
    x = np.arange(10 * 10).reshape(10, 10)
    with pytest.raises(ValueError):
        check_array(x, allow_3d=True, ensure_min_dims=2)

    x = np.arange(3 * 3 * 3).reshape(3, 3, 3)
    check_array(x, allow_3d=True, ensure_min_dims=2)


def test_check_array_allow_3d():
    x = np.arange(10 * 2 * 2 * 10).reshape(10, 2, 2, 10)
    with pytest.raises(ValueError, match="Found array with dim 4. None expected <= 3."):
        check_array(x, allow_3d=True)


def test_check_array_allow_nd():
    x = np.arange(10 * 2 * 2 * 10).reshape(10, 2, 2, 10)
    check_array(x, allow_nd=True)

    x = np.arange(10)
    with pytest.raises(ValueError):
        check_array(x, allow_nd=True)


def test_check_array_order():
    x = np.arange(10 * 3 * 10).reshape(10, 3, 10, order="f")
    x_checked = check_array(x, allow_3d=True)
    assert x_checked.flags.carray

    x_checked = check_array(x, allow_3d=True, order=None)
    assert not x_checked.flags.carray


def test_check_array_ensure2d():
    x = np.arange(10)
    check_array(x, ensure_2d=False)

    x = np.arange(10 * 10).reshape(10, 10)
    check_array(x, ensure_2d=False)


def test_check_array_ravel_1d():
    x = np.arange(10)
    assert check_array(x.reshape(10, 1), ravel_1d=True, ensure_2d=False).ndim == 1
    assert check_array(x.reshape(10, 1), ensure_2d=False).ndim == 2
    assert check_array(x, ensure_2d=False).ndim == 1

    with pytest.raises(ValueError):
        check_array(x.reshape(5, 2), ravel_1d=True)


def test_check_array_force_all_finite():
    with pytest.raises(ValueError, match=".* NaN.*"):
        check_array(
            [1, 2, 3, np.inf, np.nan, -np.inf],
            ensure_2d=False,
            force_all_finite=True,
            allow_eos=False,
        )

    with pytest.raises(ValueError, match=".*infinity.*"):
        check_array(
            [1, 2, 3, np.inf, np.nan, -np.inf],
            ensure_2d=False,
            force_all_finite="allow-nan",
            allow_eos=True,
        )

    check_array(
        [1, 2, 3, np.inf, np.nan, -np.inf], ensure_2d=False, force_all_finite=False
    )


def test_check_array_allow_eos():
    with pytest.raises(ValueError, match=".*equal length.*"):
        check_array([[1, 2, 3, 4, eos]], allow_eos=False)

    check_array([[1, np.nan, 3, eos, 10]], force_all_finite="allow-nan", allow_eos=True)

    with pytest.raises(ValueError, match=".*NaN.*"):
        check_array([[1, eos, eos, np.nan, 1]], allow_eos=True)



# Authors: Isak Samsten
# License: BSD 3 clause

"""
Symbolic aggregate approximation and piecewice aggregate approximation.
"""

import abc
import numbers

import numpy as np
from scipy.stats import norm, uniform
from sklearn.base import TransformerMixin, check_is_fitted
from sklearn.utils._param_validation import Interval, StrOptions
from sklearn.utils.validation import check_scalar

from wildboar.utils.validation import check_array, check_option

from ..base import BaseEstimator
from ._interval import IntervalTransform


def _percentiles(n_bins):
    return np.linspace(0, 1, num=n_bins, endpoint=False)[1:].reshape(1, -1)


class Binning(metaclass=abc.ABCMeta):
    """
    Abstract base class for binning strategies.

    Parameters
    ----------
    n_bins : int
        The number of bins.
    """

    def __init__(self, n_bins) -> None:
        self.n_bins = n_bins
        self._percentiles = _percentiles(n_bins)

    @abc.abstractmethod
    def scale(self, x):
        """
        Scale the input such that it is suitable for binning.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timesteps)
            The samples.
        """
        pass

    def get_thresholds(self, x=None, estimate=False):
        """
        Get thresholds for binning.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timestep), optional
            The samples.
        estimate : bool, optional
            Calculate the thresholds from the samples.

        Returns
        -------
        ndarray of shape (n_bins - 1, )
            The thresholds.
        """
        if estimate and x is None:
            raise ValueError("if estimate=True, x cannot be None.")

        return self._get_thresholds(x, estimate)

    @abc.abstractmethod
    def _get_thresholds(self, x):
        pass


class NormalBinning(Binning):
    """Bin using a normal distribution."""

    def scale(self, x):
        # Avoid circular import
        from ..datasets.preprocess import standardize

        return standardize(x)

    def _get_thresholds(self, x, estimate):
        if estimate and x is not None:
            loc = np.mean(x, axis=1).reshape(-1, 1)
            scale = np.std(x, axis=1).reshape(-1, 1)
            return norm.ppf(self._percentiles, loc=loc, scale=scale)
        else:
            ppf = norm.ppf(self._percentiles)
            if x is None:
                return ppf.reshape(-1)
            else:
                return np.repeat(ppf, x.shape[0], axis=0)


class UniformBinning(Binning):
    """Bin using a uniform distribution."""

    def scale(self, x):
        # Avoid circular import
        from ..datasets.preprocess import minmax_scale

        return minmax_scale(x)

    def _get_thresholds(self, x, estimate):
        if estimate and x is not None:
            loc = np.min(x, axis=1).reshape(-1, 1)
            scale = np.max(x, axis=1).reshape(-1, 1) - loc
            return uniform.ppf(self._percentiles, loc=loc, scale=scale)
        else:
            ppf = uniform.ppf(self._percentiles)
            if x is None:
                return ppf.reshape(-1)
            else:
                return np.repeat(ppf, x.shape[0], axis=0)


_BINNING = {"normal": NormalBinning, "uniform": UniformBinning}


class SAX(TransformerMixin, BaseEstimator):
    """
    Symbolic aggregate approximation.

    Parameters
    ----------
    n_intervals : str, optional
        The number of intervals to use for the transform.

        - if "log2", the number of intervals is `log2(n_timestep)`.
        - if "sqrt", the number of intervals is `sqrt(n_timestep)`.
        - if int, the number of intervals is `n_intervals`.
        - if float, the number of intervals is `n_intervals * n_timestep`, with
            `0 < n_intervals < 1`.
    window : int, optional
        The window size. If `window` is set, the value of `n_intervals` has no
        effect.
    n_bins : int, optional
        The number of bins.
    binning : str, optional
        The bin construction. By default the bins are defined according to the
        normal distribution. Possible values are "normal" for normally
        distributed bins or "uniform" for uniformly distributed bins.
    estimate : bool, optional
        Estimate the distribution parameters for the binning from data.

        If `estimate=False`, it is assumed that each time series is
        preprocessed using:

        - :func:`datasets.preprocess.normalize` when `binning="normal"`.
        - :func:`datasets.preprocess.minmax_scale`. when `binning="uniform"`.
    """

    _parameter_constraints: dict = {
        "n_intervals": [
            Interval(numbers.Integral, 1, None, closed="left"),
            Interval(numbers.Real, 0, 1, closed="right"),
            StrOptions({"sqrt", "log2"}),
        ],
        "window": [None, Interval(numbers.Integral, 1, None, closed="left")],
        "n_bins": [Interval(numbers.Integral, 1, None, closed="left")],
        "binning": [StrOptions({"normal", "uniform"})],
        "estimate": ["boolean"],
    }

    def __init__(
        self,
        *,
        n_intervals="sqrt",
        window=None,
        n_bins=4,
        binning="normal",
        estimate=True,
    ):
        self.n_intervals = n_intervals
        self.window = window
        self.n_bins = n_bins
        self.binning = binning
        self.estimate = estimate

    def fit(self, x, y=None):
        self._validate_params()
        x = self._validate_data(x, dtype=float)
        self.binning_ = check_option(_BINNING, self.binning, "binning")(self.n_bins)
        self.bins_ = np.arange(self.n_bins, dtype=np.min_scalar_type(self.n_bins))
        self.paa_ = PAA(n_intervals=self.n_intervals, window=self.window).fit(x)
        return self

    def transform(self, x):
        check_is_fitted(self)
        x = self._validate_data(x, reset=False, dtype=float)
        x_paa = self.paa_.transform(x)
        thresholds = self.binning_.get_thresholds(x, estimate=self.estimate)
        x_out = np.empty(x_paa.shape, dtype=np.min_scalar_type(self.n_bins))
        for i, (sample, threshold) in enumerate(zip(x_paa, thresholds)):
            x_out[i] = np.digitize(sample, threshold)
        return x_out

    def inverse_transform(self, x):
        check_is_fitted(self)
        if self.estimate:
            raise ValueError("Unable to inverse_transform with estimate=True")

        x = check_array(x, dtype=np.min_scalar_type(self.n_bins))
        thresholds = self.binning_.get_thresholds()
        thresholds = np.hstack(
            [thresholds[0], (thresholds[:-1] + thresholds[1:]) / 2, thresholds[-1]]
        )

        x_inverse = np.empty((x.shape[0], self.n_timesteps_in_), dtype=float)
        for i, (start, end) in enumerate(self.intervals):
            x_inverse[:, start:end] = thresholds[x[:, i], np.newaxis]

        return x_inverse

    @property
    def intervals(self):
        return self.paa_.intervals

    def _more_tags(self):
        return {
            "preserves_dtype": [],
        }


class PAA(TransformerMixin, BaseEstimator):
    """
    Peicewise aggregate approximation.

    Parameters
    ----------
    n_intervals : {"sqrt", "log2"}, int or float, optional
        The number of intervals.
    window : int, optional
        The size of an interval. If `window`, is given then `n_intervals` is ignored.
    """

    _parameter_constraints: dict = {
        "n_intervals": [
            Interval(numbers.Integral, 1, None, closed="left"),
            Interval(numbers.Real, 0, 1, closed="right"),
            StrOptions({"sqrt", "log2"}),
        ],
        "window": [None, Interval(numbers.Integral, 1, None, closed="left")],
    }

    def __init__(self, n_intervals="sqrt", window=None):
        self.n_intervals = n_intervals
        self.window = window

    def fit(self, x, y=None):
        self._validate_params()
        x = self._validate_data(x, dtype=float)
        if self.window is not None:
            n_intervals = x.shape[-1] // check_scalar(
                self.window,
                "window",
                numbers.Integral,
                min_val=1,
                max_val=self.n_timesteps_in_,
            )
        else:
            n_intervals = self.n_intervals

        self.interval_transform_ = IntervalTransform(
            n_intervals=n_intervals, summarizer="mean"
        )
        self.interval_transform_.fit(x)
        return self

    def transform(self, x):
        check_is_fitted(self)
        x = self._validate_data(x, dtype=float, reset=False)
        return self.interval_transform_.transform(x)

    def inverse_transform(self, x):
        check_is_fitted(self)
        x = check_array(x, dtype=float)

        x_inverse = np.empty((x.shape[0], self.n_timesteps_in_), dtype=float)
        for i, (start, end) in enumerate(self.intervals):
            x_inverse[:, start:end] = x[:, i].reshape(-1, 1)

        return x_inverse

    @property
    def intervals(self):
        return [
            (start, start + length)
            for (
                _,
                (start, length, _),
            ) in self.interval_transform_.embedding_.attributes
        ]


def symbolic_aggregate_approximation(
    x,
    *,
    n_intervals="sqrt",
    window=None,
    n_bins=4,
    binning="normal",
):
    """
    Symbolic aggregate approximation.

    Parameters
    ----------
    x : array-like of shape (n_samples, n_timestep)
        The input data.
    n_intervals : str, optional
        The number of intervals to use for the transform.

        - if "log2", the number of intervals is ``log2(n_timestep)``.
        - if "sqrt", the number of intervals is ``sqrt(n_timestep)``.
        - if int, the number of intervals is ``n_intervals``.
        - if float, the number of intervals is ``n_intervals * n_timestep``, with
            ``0 < n_intervals < 1``.
    window : int, optional
        The window size. If ``window`` is set, the value of ``n_intervals`` has no
        effect.
    n_bins : int, optional
        The number of bins.
    binning : str, optional
        The bin construction. By default the bins are defined according to the
        normal distribution. Possible values are ``"normal"`` for normally
        distributed bins or ``"uniform"`` for uniformly distributed bins.

    Returns
    -------
    ndarray of shape (n_samples, n_intervals)
        The symbolic aggregate approximation.

    """
    return SAX(
        n_intervals=n_intervals, window=window, n_bins=n_bins, binning=binning
    ).fit_transform(x)


def piecewice_aggregate_approximation(x, *, n_intervals="sqrt", window=None):
    """
    Peicewise aggregate approximation.

    Parameters
    ----------
    x : array-like of shape (n_samples, n_timestep)
        The input data.
    n_intervals : str, optional
        The number of intervals to use for the transform.

        - if "log2", the number of intervals is ``log2(n_timestep)``.
        - if "sqrt", the number of intervals is ``sqrt(n_timestep)``.
        - if int, the number of intervals is ``n_intervals``.
        - if float, the number of intervals is ``n_intervals * n_timestep``, with
            ``0 < n_intervals < 1``.
    window : int, optional
        The window size. If ``window`` is set, the value of ``n_intervals`` has no
        effect.

    Returns
    -------
    ndarray of shape (n_samples, n_intervals)
        The symbolic aggregate approximation.

    """
    return PAA(n_intervals=n_intervals, window=window).fit_transform(x)



import numpy as np
import pytest

from wildboar.utils._misc import _test_ts_array
from wildboar.utils.validation import _check_ts_array


def test_check_tsarray_skip_2D():
    X = np.arange(10 * 10).reshape(10, 10)
    X = X[::2, :]
    X_checked = _check_ts_array(X)
    assert X.base is not X_checked.base


def test_check_tsarray_skip_3D():
    X = np.arange(10 * 3 * 10).reshape(10, 3, 10)
    X = X[::2, :, :]
    X_checked = _check_ts_array(X)
    assert X.base is not X_checked.base
    assert X_checked.dtype == float


def test_check_tsarray_slice_3D():
    X = np.zeros((10, 4, 10))
    X_sliced = X[::2, ::2, :4:2]
    assert 0 == _test_ts_array(_check_ts_array(X_sliced))

    with pytest.raises(
        ValueError,
        match="Buffer and memoryview are not contiguous in the same dimension",
    ):
        _test_ts_array(X_sliced)


def test_check_tsarray_slice_same_base_3D():
    X = np.zeros((10, 3, 10))
    X_sliced = X[::2, :, :3]
    assert _check_ts_array(X_sliced).base is X

    X_sliced = X[::2, ::2, ::2]
    assert _check_ts_array(X_sliced).base is not X



# Authors: Isak Samsten
# License: BSD 3 clause

import math
import numbers

import numpy as np
from sklearn.utils._param_validation import Interval, StrOptions

from ..distance._distance import _METRICS, _SUBSEQUENCE_METRICS
from ..distance._multi_metric import make_subsequence_metrics
from ._base import BaseAttributeTransform
from ._cshapelet import (
    CastorAttributeGenerator,
    CastorSummarizer,
    DilatedShapeletAttributeGenerator,
    RandomMultiMetricShapeletAttributeGenerator,
    RandomShapeletAttributeGenerator,
)


class ShapeletMixin:
    """Mixin for shapelet based estiomators."""

    _parameter_constraints: dict = {
        "n_shapelets": [
            Interval(numbers.Integral, 1, None, closed="left"),
            StrOptions({"log2", "sqrt"}),
            callable,
        ],
        "metric": [
            StrOptions(_SUBSEQUENCE_METRICS.keys()),
            list,
        ],
        "metric_params": [dict, None],
        "min_shapelet_size": [
            Interval(numbers.Real, 0, 1, closed="both"),
        ],
        "max_shapelet_size": [
            Interval(numbers.Real, 0, 1, closed="both"),
        ],
    }

    def _get_generator(self, x, y):  # noqa: PLR0912
        if self.min_shapelet_size > self.max_shapelet_size:
            raise ValueError(
                f"The min_shapelet_size parameter of {type(self).__qualname__} "
                "must be <= max_shapelet_size."
            )

        max_shapelet_size = math.ceil(self.n_timesteps_in_ * self.max_shapelet_size)
        min_shapelet_size = math.ceil(self.n_timesteps_in_ * self.min_shapelet_size)
        if min_shapelet_size < 2:
            # NOTE: To ensure that the same random_seed generates the same shapelets
            # in future versions we keep the limit of 2 timesteps for a shapelet as long
            # as the time series is at least 2 timesteps. Otherwise we fall back to 1
            # timestep.
            if self.n_timesteps_in_ < 2:
                min_shapelet_size = 1
            else:
                min_shapelet_size = 2

        if isinstance(self.n_shapelets, str) or callable(self.n_shapelets):
            if min_shapelet_size < max_shapelet_size:
                possible_shapelets = sum(
                    self.n_timesteps_in_ - curr_len + 1
                    for curr_len in range(min_shapelet_size, max_shapelet_size)
                )
            else:
                possible_shapelets = self.n_timesteps_in_ - min_shapelet_size + 1

            if self.n_shapelets == "log2":
                n_shapelets = int(np.log2(possible_shapelets))
            elif self.n_shapelets == "sqrt":
                n_shapelets = int(np.sqrt(possible_shapelets))
            else:
                n_shapelets = int(self.n_shapelets(possible_shapelets))
        else:
            n_shapelets = self.n_shapelets

        if isinstance(self.metric, str):
            metric_params = self.metric_params if self.metric_params is not None else {}
            return RandomShapeletAttributeGenerator(
                _SUBSEQUENCE_METRICS[self.metric](**metric_params),
                min_shapelet_size,
                max_shapelet_size,
                max(1, n_shapelets),
            )
        else:
            metrics, weights = make_subsequence_metrics(self.metric)
            return RandomMultiMetricShapeletAttributeGenerator(
                max(1, n_shapelets),
                min_shapelet_size,
                max_shapelet_size,
                metrics,
                weights,
            )


def _odd_shapelet_size(shapelet_size, n_timesteps):
    if shapelet_size == 1:
        shapelet_size = 3

    if shapelet_size % 2 == 0:
        shapelet_size += 1
    return shapelet_size


class DilatedShapeletMixin:
    _parameter_constraints = {
        "n_shapelets": [Interval(numbers.Integral, 1, None, closed="left")],
        "metric_params": [dict, None],
        "metric": [StrOptions(_METRICS.keys())],
        "min_shapelet_size": [None, Interval(numbers.Real, 0, 1, closed="both")],
        "max_shapelet_size": [None, Interval(numbers.Real, 0, 1, closed="both")],
        "shapelet_size": [None, "array-like"],
        "normalize_prob": [Interval(numbers.Real, 0, 1, closed="both")],
        "lower": [Interval(numbers.Real, 0, 1, closed="both")],
        "upper": [Interval(numbers.Real, 0, 1, closed="both")],
        "ignore_y": ["boolean"],
    }

    def _get_generator(self, x, y):
        metric_params = self.metric_params if self.metric_params is not None else {}
        Metric = _METRICS[self.metric]

        if self.min_shapelet_size is not None or self.max_shapelet_size:
            if self.shapelet_size is not None:
                raise ValueError(
                    "Both shapelet_size and min_shapelet_size or max_shapelet_size "
                    "cannot be set at the same time"
                )

            min_shapelet_size = (
                self.min_shapelet_size if self.min_shapelet_size is not None else 0
            )
            max_shapelet_size = (
                self.max_shapelet_size if self.max_shapelet_size is not None else 1
            )
            if min_shapelet_size > max_shapelet_size:
                raise ValueError(
                    "max_shapelet_size must be larger than min_shapelet_size"
                )

            min_shapelet_size = int(self.n_timesteps_in_ * min_shapelet_size)
            max_shapelet_size = int(self.n_timesteps_in_ * max_shapelet_size)
            if min_shapelet_size < 2:
                min_shapelet_size = 2
            if max_shapelet_size < 3:
                max_shapelet_size = 3
            shapelet_size = range(min_shapelet_size, max_shapelet_size)
        elif self.shapelet_size is None:
            shapelet_size = [7, 9, 11]
        else:
            shapelet_size = self.shapelet_size

        shapelet_size = np.array(
            [_odd_shapelet_size(size, x.shape[-1]) for size in shapelet_size]
        )

        if self.lower > self.upper:
            raise ValueError("Lower can't be larger than upper")

        if y is not None and (not hasattr(self, "ignore_y") or not self.ignore_y):
            _, y, samples_per_label = np.unique(
                np.array(y), return_inverse=True, return_counts=True
            )
            samples = np.argsort(y)
        else:
            samples = None
            samples_per_label = None
            y = None

        return DilatedShapeletAttributeGenerator(
            Metric(**metric_params),
            self.n_shapelets,
            shapelet_size,
            self.normalize_prob,
            self.lower,
            self.upper,
            y,
            samples,
            samples_per_label,
        )


class CastorMixin:
    _parameter_constraints = {
        "n_groups": [Interval(numbers.Integral, 1, None, closed="left")],
        "n_shapelets": [Interval(numbers.Integral, 1, None, closed="left")],
        "metric_params": [dict, None],
        "metric": [StrOptions(_METRICS.keys())],
        "shapelet_size": [Interval(numbers.Integral, 3, None, closed="left")],
        "normalize_prob": [Interval(numbers.Real, 0, 1, closed="both")],
        "lower": [Interval(numbers.Real, 0, 1, closed="both")],
        "upper": [Interval(numbers.Real, 0, 1, closed="both")],
        "soft_min": [bool],
        "soft_max": [bool],
        "soft_threshold": [bool],
        "ignore_y": ["boolean"],
    }

    def _get_generator(self, x, y):
        Metric = _METRICS[self.metric]
        metric_params = self.metric_params if self.metric_params is not None else {}
        if y is not None and (not hasattr(self, "ignore_y") or not self.ignore_y):
            _, y, samples_per_label = np.unique(
                np.array(y), return_inverse=True, return_counts=True
            )
            samples = np.argsort(y)
        else:
            samples = None
            samples_per_label = None
            y = None

        return CastorAttributeGenerator(
            self.n_groups,
            self.n_shapelets,
            _odd_shapelet_size(self.shapelet_size, x.shape[-1]),
            self.normalize_prob,
            self.lower,
            self.upper,
            Metric(**metric_params),
            CastorSummarizer(self.soft_min, self.soft_max, self.soft_threshold),
            y,
            samples,
            samples_per_label,
        )


class CastorTransform(CastorMixin, BaseAttributeTransform):
    """
    Competing Dialated Shapelet Transform.

    Parameters
    ----------
    n_groups : int, optional
        The number of groups of dilated shapelets.
    n_shapelets : int, optional
        The number of dilated shapelets per group.
    metric : str or callable, optional
        The distance metric

        See ``_METRICS.keys()`` for a list of supported metrics.
    metric_params : dict, optional
        Parameters to the metric.

        Read more about the parameters in the
        :ref:`User guide <list_of_metrics>`.
    normalize_prob : float, optional
        The probability of standardizing a shapelet with zero mean and unit
        standard deviation.
    shapelet_size : int, optional
        The length of the dilated shapelet.
    lower : float, optional
        The lower percentile to draw distance thresholds above.
    upper : float, optional
        The upper percentile to draw distance thresholds below.
    ignore_y : bool, optional
        Ignore `y` and use the same sample which a shapelet is sampled from to
        estimate the distance threshold.
    random_state : int or RandomState, optional
        Controls the random sampling of kernels.

        - If `int`, `random_state` is the seed used by the random number
          generator.
        - If :class:`numpy.random.RandomState` instance, `random_state` is
          the random number generator.
        - If `None`, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.
    n_jobs : int, optional
        The number of parallel jobs.
    """

    _parameter_constraints = {
        **CastorMixin._parameter_constraints,
        **BaseAttributeTransform._parameter_constraints,
    }

    def __init__(
        self,
        n_groups=64,
        n_shapelets=8,
        *,
        metric="euclidean",
        metric_params=None,
        normalize_prob=0.8,
        shapelet_size=11,
        lower=0.05,
        upper=0.1,
        soft_min=True,
        soft_max=False,
        soft_threshold=True,
        ignore_y=False,
        random_state=None,
        n_jobs=None,
    ):
        super().__init__(n_jobs=n_jobs, random_state=random_state)
        self.n_groups = n_groups
        self.n_shapelets = n_shapelets
        self.metric = metric
        self.metric_params = metric_params
        self.normalize_prob = normalize_prob
        self.shapelet_size = shapelet_size
        self.lower = lower
        self.upper = upper
        self.soft_min = soft_min
        self.soft_max = soft_max
        self.soft_threshold = soft_threshold
        self.ignore_y = ignore_y


class DilatedShapeletTransform(DilatedShapeletMixin, BaseAttributeTransform):
    """
    Dilated shapelet transform.

    Transform time series to a representation consisting of three
    values per shapelet: minimum dilated distance, the index of
    the timestep that minimizes the distance and number of subsequences
    that are below a distance threshold.

    Parameters
    ----------
    n_shapelets : int, optional
        The number of dilated shapelets.
    metric : str or callable, optional
        The distance metric

        See ``_METRICS.keys()`` for a list of supported metrics.
    metric_params : dict, optional
        Parameters to the metric.

        Read more about the parameters in the
        :ref:`User guide <list_of_metrics>`.
    normalize_prob : float, optional
        The probability of standardizing a shapelet with zero mean and unit
        standard deviation.
    min_shapelet_size : float, optional
        The minimum shapelet size. If None, use the discrete sizes
        in `shapelet_size`.
    max_shapelet_size : float, optional
        The maximum shapelet size. If None, use the discrete sizes
        in `shapelet_size`.
    shapelet_size : array-like, optional
        The size of shapelets, by default [7, 9, 11].
    lower : float, optional
        The lower percentile to draw distance thresholds above.
    upper : float, optional
        The upper percentile to draw distance thresholds below.
    ignore_y : bool, optional
        Ignore `y` and use the same sample which a shapelet is sampled from to
        estimate the distance threshold.
    random_state : int or RandomState, optional
        Controls the random sampling of kernels.

        - If `int`, `random_state` is the seed used by the random number
          generator.
        - If :class:`numpy.random.RandomState` instance, `random_state` is
          the random number generator.
        - If `None`, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.
    n_jobs : int, optional
        The number of parallel jobs.

    References
    ----------
    Antoine Guillaume, Christel Vrain, Elloumi Wael
        Random Dilated Shapelet Transform: A New Approach for Time Series Shapelets
        Pattern Recognition and Artificial Intelligence, 2022
    """

    _parameter_constraints = {
        **DilatedShapeletMixin._parameter_constraints,
        **BaseAttributeTransform._parameter_constraints,
    }

    def __init__(
        self,
        n_shapelets=1000,
        *,
        metric="euclidean",
        metric_params=None,
        normalize_prob=0.5,
        min_shapelet_size=None,
        max_shapelet_size=None,
        shapelet_size=None,
        lower=0.05,
        upper=0.1,
        ignore_y=False,
        random_state=None,
        n_jobs=None,
    ):
        super().__init__(random_state=random_state, n_jobs=n_jobs)
        self.n_shapelets = n_shapelets
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.shapelet_size = shapelet_size
        self.metric = metric
        self.metric_params = metric_params
        self.normalize_prob = normalize_prob
        self.lower = lower
        self.upper = upper
        self.ignore_y = ignore_y


class RandomShapeletTransform(ShapeletMixin, BaseAttributeTransform):
    """
    Random shapelet tranform.

    Transform a time series to the distances to a selection of random
    shapelets.

    Parameters
    ----------
    n_shapelets : int, optional
        The number of shapelets in the resulting transform.
    metric : str or list, optional
        - If str, the distance metric used to identify the best shapelet.
        - If list, multiple metrics specified as a list of tuples, where the first
            element of the tuple is a metric name and the second element a dictionary
            with a parameter grid specification. A parameter grid specification is a
            dict with two mandatory and one optional key-value pairs defining the
            lower and upper bound on the values and number of values in the grid. For
            example, to specifiy a grid over the argument 'r' with 10 values in the
            range 0 to 1, we would give the following specification: ``dict(min_r=0,
            max_r=1, num_r=10)``.

        Read more about the metrics and their parameters in the
        :ref:`User guide <list_of_subsequence_metrics>`.
    metric_params : dict, optional
        Parameters for the distance measure. Ignored unless metric is a string.

        Read more about the parameters in the :ref:`User guide
        <list_of_subsequence_metrics>`.
    min_shapelet_size : float, optional
        Minimum shapelet size.
    max_shapelet_size : float, optional
        Maximum shapelet size.
    n_jobs : int, optional
        The number of jobs to run in parallel. None means 1 and -1 means using all
        processors.
    random_state : int or RandomState, optional
        - If `int`, `random_state` is the seed used by the random number generator
        - If `RandomState` instance, `random_state` is the random number generator
        - If `None`, the random number generator is the `RandomState` instance used
            by `np.random`.

    Attributes
    ----------
    embedding_ : Embedding
        The underlying embedding object.

    References
    ----------
    Wistuba, Martin, Josif Grabocka, and Lars Schmidt-Thieme.
        Ultra-fast shapelets for time series classification. arXiv preprint
        arXiv:1503.05018 (2015).

    Examples
    --------
    Transform each time series to the minimum DTW distance to each shapelet

    >>> from wildboar.dataset import load_gunpoint()
    >>> from wildboar.transform import RandomShapeletTransform
    >>> t = RandomShapeletTransform(metric="dtw")
    >>> t.fit_transform(X)

    Transform each time series to the either the minimum DTW distance, with r randomly
    set set between 0 and 1 or ERP distance with g between 0 and 1.

    >>> t = RandomShapeletTransform(
    ...     metric=[
    ...         ("dtw", dict(min_r=0.0, max_r=1.0)),
    ...         ("erp", dict(min_g=0.0, max_g=1.0)),
    ...     ]
    ... )
    >>> t.fit_transform(X)
    """

    _parameter_constraints: dict = {
        **ShapeletMixin._parameter_constraints,
        **BaseAttributeTransform._parameter_constraints,
    }

    def __init__(
        self,
        n_shapelets=1000,
        *,
        metric="euclidean",
        metric_params=None,
        min_shapelet_size=0.0,
        max_shapelet_size=1.0,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(n_jobs=n_jobs, random_state=random_state)
        self.n_shapelets = n_shapelets
        self.metric = metric
        self.metric_params = metric_params
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
